import numpy as np
import os
from TB2J.utils import symbol_number
from numpy import array_str


def write_info_section(cls, myfile):
    myfile.write("=" * 90 + "\n")
    myfile.write("Information: \n")
    # myfile.write("Exchange parameters generated by TB2J %s\n" % (__version__))
    # myfile.write("Generation time: %s\n" % (now.strftime("%y/%m/%d %H:%M:%S")))
    myfile.write(cls.description)


def write_atom_section(cls, myfile):
    """
    write the atom section
    including the cell and the atomic positions.
    """
    myfile.write("=" * 90 + "\n")
    myfile.write("Cell (Angstrom):\n")
    cell = cls.atoms.get_cell()
    for c in cell:
        myfile.write("{:10.8f}  {:10.8f}  {:10.8f}\n".format(c[0], c[1], c[2]))

    # write atom information
    myfile.write("\n")
    myfile.write("=" * 90 + "\n")
    myfile.write("Atoms:  \n")
    myfile.write("(Note: charge and magmoms only count the wannier functions.)\n")
    if cls.colinear:
        myfile.write(
            "{:^12s} {:^13s} {:^13s} {:^13s} {:^9s} {:^9s}\n".format(
                "Atom number", "x", "y", "z", "w_charge", "w_magmom"
            )
        )
    else:
        myfile.write(
            "{:^12s} {:^13s} {:^13s} {:^13s} {:^9s} {:^9s} {:^9s} {:^9s}\n".format(
                "Atom_number", "x", "y", "z", "w_charge", "M(x)", "M(y)", "M(z)"
            )
        )

    symnum = symbol_number(cls.atoms)
    sns = list(symnum.keys())
    poses = cls.atoms.get_positions()

    tchg, tmx, tmy, tmz = 0, 0, 0, 0
    for i, s in enumerate(symnum):
        if cls.colinear:
            mag = cls.magmoms[i]
            chg = cls.charges[i]
            myfile.write(
                "{:<12s} {:13.8f} {:13.8f} {:13.8f} {:9.4f} {:9.4f}\n".format(
                    s, poses[i, 0], poses[i, 1], poses[i, 2], chg, mag
                )
            )
        else:
            chg = cls.charges[i]
            mx, my, mz = cls.spinat[i, :]
            myfile.write(
                "{:<12s} {:13.8f} {:13.8f} {:13.8f} {:9.4f} {:9.4f} {:9.4f} {:9.4f}\n".format(
                    s, poses[i, 0], poses[i, 1], poses[i, 2], chg, mx, my, mz
                )
            )
            tchg += chg
            tmx += mx
            tmy += my
            tmz += mz
    if cls.colinear:
        myfile.write(
            "{:<12s} {:13s} {:13s} {:13s} {:9.4f} {:9.4f}\n".format(
                "Total", "", "", "", np.sum(cls.charges), np.sum(cls.magmoms)
            )
        )
    else:
        myfile.write(
            "{:<12s} {:13s} {:13s} {:13s} {:9.4f} {:9.4f} {:9.4f} {:9.4f}\n".format(
                "Total", "", "", "", tchg, tmx, tmy, tmz
            )
        )

    myfile.write("\n")


def write_orbital_section(cls, myfile):
    if cls.Jiso_orb:
        myfile.write("=" * 90 + "\n")
        myfile.write("Orbitals used in decomposition: \n")
        myfile.write("The name of the orbitals for the decomposition: \n")
        symnum = symbol_number(cls.atoms)
        sns = list(symnum.keys())
        for iatom in cls.orbital_names:
            if cls.index_spin[iatom] != -1:
                myfile.write(f"{sns[iatom]} : {cls.orbital_names[iatom]}\n")
        myfile.write("\n")


def write_exchange_section(
    cls,
    myfile,
    order="distance",
    write_experimental: bool = True,
    write_orb_decomposition: bool = False,
):
    symnum = symbol_number(cls.atoms)
    sns = list(symnum.keys())
    poses = cls.atoms.get_positions()

    myfile.write("=" * 90 + "\n")
    myfile.write("Exchange: \n")

    # l = [x for x in cls.exchange_Jlist if abs(x.J) > cutoff]
    keys = cls.exchange_Jdict
    if order == "amplitude":
        l = sorted(keys, key=lambda x: abs(cls.exchange_Jdict[x]), reverse=True)
    elif order == "distance":
        l = sorted(keys, key=lambda x: cls.distance_dict[x][1])
    else:
        l = keys

    myfile.write(
        "{:6s} {:5s} {:15s} {:7s} {:24s} {:11s} \n".format(
            "    i",
            "    j",
            "         R",
            "  J_iso(meV)",
            "         vector",
            "distance(A)",
        )
    )

    for ll in l:
        myfile.write("-" * 88 + "\n")
        R, i, j = ll
        J = cls.exchange_Jdict[ll]
        vec, distance = cls.distance_dict[ll]
        myfile.write(
            "   {:5s} {:5s} ({:3d}, {:3d}, {:3d}) {:7.4f}   ({:6.3f}, {:6.3f}, {:6.3f}) {:6.3f} \n".format(
                sns[cls.ind_atoms[i]],
                sns[cls.ind_atoms[j]],
                R[0],
                R[1],
                R[2],
                J * 1e3,
                vec[0],
                vec[1],
                vec[2],
                distance,
            )
        )

        Jiso = cls.exchange_Jdict[ll] * 1e3
        myfile.write(f"J_iso: {Jiso:7.4f} \n")

        write_experimental = False

        if cls.has_biquadratic and write_experimental:
            Jprime, B = cls.biquadratic_Jdict[ll]
            myfile.write(f"[Testing!] Jprime: {Jprime*1e3:.3f},  B: {B*1e3:.3f}\n")

        if cls.dJdx is not None:
            dJdx = cls.dJdx[ll]
            myfile.write(f"dJ/dx: {dJdx*1e3:.3f}\n")

        if cls.dJdx2 is not None:
            dJdx2 = cls.dJdx2[ll]
            myfile.write(f"d2J/dx2: {dJdx2*1e3:.3f}\n")

        if cls.dmi_ddict is not None:
            DMI = cls.dmi_ddict[ll] * 1e3
            myfile.write(
                "[Testing!] DMI: ({:7.4f} {:7.4f} {:7.4f})\n".format(
                    DMI[0], DMI[1], DMI[2]
                )
            )

        if write_experimental:
            try:
                DMI2 = cls.debug_dict["DMI2"][ll] * 1e3
                myfile.write(
                    "[Debug!] DMI2: ({:7.4f} {:7.4f} {:7.4f})\n".format(
                        DMI2[0], DMI2[1], DMI2[2]
                    )
                )
                pass
            except:
                pass

        if cls.Jani_dict is not None:
            J = cls.Jani_dict[ll] * 1e3
            myfile.write(
                f"[Testing!]J_ani:\n{array_str(J, precision=3, suppress_small=True)}\n"
            )

        if cls.NJT_ddict is not None:
            DMI = cls.NJT_ddict[ll] * 1e3
            myfile.write(
                "[Experimental!] DMI_NJt: ({:7.4f} {:7.4f} {:7.4f})\n".format(
                    DMI[0], DMI[1], DMI[2]
                )
            )

        if cls.NJT_Jdict is not None:
            J = cls.NJT_Jdict[ll] * 1e3
            myfile.write(
                "[Testing!] Jani_NJt: ({:7.4f} {:7.4f} {:7.4f})\n".format(
                    J[0], J[1], J[2]
                )
            )

        if write_orb_decomposition:
            if cls.Jiso_orb:
                myfile.write(
                    "Orbital contributions:\n isotropic J:\n {} \n".format(
                        np.array_str(
                            cls.Jiso_orb[ll] * 1e3, precision=3, suppress_small=True
                        )
                    )
                )

            xyz = "xyz"
            if cls.DMI_orb:
                for i in range(3):
                    myfile.write(f"DMI {xyz[i]}:\n")
                    myfile.write(
                        np.array_str(
                            cls.DMI_orb[ll][i] * 1e3, precision=3, suppress_small=True
                        )
                    )
                    myfile.write("\n")

            if cls.Jani_orb:
                for i in range(3):
                    for j in range(3):
                        myfile.write(f"Jani {xyz[i]}{xyz[j]}:\n")
                        myfile.write(
                            np.array_str(
                                cls.Jani_orb[ll][i, j] * 1e3,
                                precision=3,
                                suppress_small=True,
                            )
                        )
                        myfile.write("\n")


def to_str(arr):
    return " ".join([f"{x:8.5f}" for x in arr])


def write_Jq_info(cls, kpts, evals, evecs, myfile, special_kpoints={}):
    """
    kpts: the list of kpoints
    evals: eigen values
    evecs: eigen vectors
    fname: the name of the output file
    spk: special kpoints
    """
    symnum = symbol_number(cls.atoms)
    sns = list(symnum.keys())
    imin = np.argmin(evals[:, 0])
    # emin = np.min(evals[:, 0])
    nspin = evals.shape[1] // 3
    nband = evals.shape[1]
    evec_min = evecs[imin, :, 0].reshape(nspin, 3)

    evals = evals / 1.6e-19 * 1000

    # write qpoint info
    myfile.write("=" * 60 + "\n")
    myfile.write("List of qpoints\n")
    myfile.write(f"Number of qpoints: {len(kpts)} .\n")
    for ik, k in enumerate(kpts):
        myfile.write(f"Index={ik}: {k[0]:9.5f} {k[1]:9.5f} {k[2]:9.5f}\n")

    # write eigen values and eigen vector for each qpoint
    myfile.write("=" * 60 + "\n")
    myfile.write("Eigen values and Eigen vectors:\n\n")

    for ik, k in enumerate(kpts):
        myfile.write("-" * 59 + "\n")
        myfile.write(f"Index={ik}: {k[0]:9.5f} {k[1]:9.5f} {k[2]:9.5f}\n")
        myfile.write(f"Eigen values (meV): {to_str(evals[ik])} \n")

        for ib in range(nband):
            myfile.write(f"\nBand_index: {ib}, Eigenvalue: {evals[ik, ib]:9.5f}\n")
            myfile.write(f"Eigenstate: {to_str(evecs[ik, :, ib])}\n")

    myfile.write("=" * 60)
    myfile.write("\nThe energy minimum is at the q-point:")
    myfile.write("%s\n" % kpts[np.argmin(evals[:, 0])])
    myfile.write("The eigenstate at the energy minimum is:\n")
    for i, ev in enumerate(evec_min):
        v = ev / np.linalg.norm(ev)
        myfile.write(
            f"{sns[cls.ind_atoms[i]]}, {v[0]: 8.3f}, {v[2]: 8.3f}, {v[2]: 8.3f}\n"
        )

    print("\nThe energy minimum is at:")
    print("%s\n" % kpts[np.argmin(evals[:, 0])])
    print("The eigenstate at the energy minimum is:\n")
    for i, ev in enumerate(evec_min):
        v = ev / np.linalg.norm(ev)
        print(f"{sns[cls.ind_atoms[i]]}, {v[0]: 8.3f}, {v[2]: 8.3f}, {v[2]: 8.3f}\n")


def write_txt(
    cls,
    path="TB2J_results",
    fname="exchange.out",
    order="distance",
    write_orb_decomposition=False,
    write_experimental=True,
    cutoff=1e-4,
):
    if not os.path.exists(path):
        os.makedirs(path)
    fname = os.path.join(path, fname)
    with open(fname, "w") as myfile:
        write_info_section(cls, myfile)
        write_atom_section(cls, myfile)
        write_orbital_section(cls, myfile)
        write_exchange_section(
            cls,
            myfile,
            write_experimental=write_experimental,
            write_orb_decomposition=write_orb_decomposition,
        )



"""
Exchange from Green's function
"""

import os
from collections import defaultdict

import numpy as np
from tqdm import tqdm

from TB2J.external import p_map
from TB2J.green import TBGreen
from TB2J.io_exchange import SpinIO

from .exchange import ExchangeCL


class ExchangeCL2(ExchangeCL):
    def set_tbmodels(self, tbmodels):
        """
        only difference is a colinear tag.
        """
        self.tbmodel_up, self.tbmodel_dn = tbmodels
        self.backend_name = self.tbmodel_up.name
        self.Gup = TBGreen(
            tbmodel=self.tbmodel_up,
            kmesh=self.kmesh,
            efermi=self.efermi,
            use_cache=self._use_cache,
            nproc=self.nproc,
        )
        self.Gdn = TBGreen(
            tbmodel=self.tbmodel_dn,
            kmesh=self.kmesh,
            efermi=self.efermi,
            use_cache=self._use_cache,
            nproc=self.nproc,
        )
        if self.write_density_matrix:
            self.Gup.write_rho_R(
                Rlist=self.Rlist, fname=os.path.join(self.output_path, "rho_up.pickle")
            )
            self.Gdn.write_rho_R(
                Rlist=self.Rlist, fname=os.path.join(self.output_path, "rho_dn.pickle")
            )
        self.norb = self.Gup.norb
        self.nbasis = self.Gup.nbasis + self.Gdn.nbasis
        # self.rho_up_list = []
        # self.rho_dn_list = []
        self.rho_up = self.Gup.get_density_matrix()
        self.rho_dn = self.Gdn.get_density_matrix()
        self.Jorb_list = defaultdict(lambda: [])
        self.JJ_list = defaultdict(lambda: [])
        self.JJ = defaultdict(lambda: 0.0j)
        self.Jorb = defaultdict(lambda: 0.0j)
        self.HR0_up = self.Gup.H0
        self.HR0_dn = self.Gdn.H0
        self.Delta = self.HR0_up - self.HR0_dn
        if self.Gup.is_orthogonal and self.Gdn.is_orthogonal:
            self.is_orthogonal = True
        else:
            self.is_orthogonal = False
            # self.S0=self.Gup.S0

        self.exchange_Jdict = {}
        self.Jiso_orb = {}

        self.biquadratic = False

        self._is_collinear = True

    def _clean_tbmodels(self):
        del self.tbmodel_up
        del self.tbmodel_dn
        del self.Gup.tbmodel
        del self.Gdn.tbmodel

    def _adjust_emin(self):
        emin_up = self.Gup.find_energy_ingap(rbound=self.efermi - 5.0) - self.efermi
        emin_dn = self.Gdn.find_energy_ingap(rbound=self.efermi - 5.0) - self.efermi
        self.emin = min(emin_up, emin_dn)
        print(f"A gap is found at {self.emin}, set emin to it.")

    def get_Delta(self, iatom):
        orbs = self.iorb(iatom)
        return self.Delta[np.ix_(orbs, orbs)]
        # s = self.orb_slice[iatom]
        # return self.Delta[s, s]

    def GR_atom(self, GR, iatom, jatom):
        """Given a green's function matrix, return the [iatom, jatom] component.

        :param GR:  Green's function matrix
        :param iatom:  index of atom i
        :param jatom:  index of atom j
        :returns:  G_ij
        :rtype:  complex matrix.
        """
        orbi = self.iorb(iatom)
        orbj = self.iorb(jatom)
        return GR[np.ix_(orbi, orbj)]
        # return GR[self.orb_slice[iatom], self.orb_slice[jatom]]

    def get_A_ijR(self, Gup, Gdn, iatom, jatom):
        """
        ! Note: not used. In get_all_A, it is reimplemented.
        """
        Rij_done = set()
        Jorb_list = dict()
        JJ_list = dict()
        for R, ijpairs in self.R_ijatom_dict.items():
            if (iatom, jatom) in ijpairs and (R, iatom, jatom) not in Rij_done:
                Gij_up = self.GR_atom(Gup[R], iatom, jatom)
                Rm = tuple(-x for x in R)
                Gji_dn = self.GR_atom(Gdn[Rm], jatom, iatom)
                tmp = 0.0j
                Deltai = self.get_Delta(iatom)
                Deltaj = self.get_Delta(jatom)
                t = np.einsum(
                    "ij, ji-> ij", np.matmul(Deltai, Gij_up), np.matmul(Deltaj, Gji_dn)
                )

                # if self.biquadratic:
                #    A = np.einsum(
                #        "ij, ji-> ij",
                #        np.matmul(Deltai, Gij_up),
                #        np.matmul(Deltaj, Gji_up),
                #    )
                #    C = np.einsum(
                #        "ij, ji-> ij",
                #        np.matmul(Deltai, Gij_down),
                #        np.matmul(Deltaj, Gji_down),
                #    )
                tmp = np.sum(t)
                self.Jorb_list[(R, iatom, jatom)].append(t / (4.0 * np.pi))
                self.JJ_list[(R, iatom, jatom)].append(tmp / (4.0 * np.pi))
                Rij_done.add((R, iatom, jatom))
                if (Rm, jatom, iatom) not in Rij_done:
                    Jorb_list[(Rm, jatom, iatom)] = t.T / (4.0 * np.pi)
                    JJ_list[(Rm, jatom, iatom)] = tmp / (4.0 * np.pi)
                    Rij_done.add((Rm, jatom, iatom))
        return Jorb_list, JJ_list

    def get_all_A(self, Gup, Gdn):
        """
        Calculate all A matrix elements
        Loop over all magnetic atoms.
        :param G: Green's function.
        :param de: energy step.
        """
        Rij_done = set()
        Jorb_list = dict()
        JJ_list = dict()
        for R, ijpairs in self.R_ijatom_dict.items():
            for iatom, jatom in ijpairs:
                if (R, iatom, jatom) not in Rij_done:
                    Rm = tuple(-x for x in R)
                    if (Rm, jatom, iatom) in Rij_done:
                        raise KeyError(
                            f"Strange (Rm, jatom, iatom) has already been calculated! {(Rm, jatom, iatom)}"
                        )
                    Gij_up = self.GR_atom(Gup[R], iatom, jatom)
                    Gji_dn = self.GR_atom(Gdn[Rm], jatom, iatom)
                    tmp = 0.0j
                    # t = self.get_Delta(iatom) @ Gij_up @ self.get_Delta(jatom) @ Gji_dn
                    t = np.einsum(
                        "ij, ji-> ij",
                        np.matmul(self.get_Delta(iatom), Gij_up),
                        np.matmul(self.get_Delta(jatom), Gji_dn),
                    )
                    tmp = np.sum(t)
                    Jorb_list[(R, iatom, jatom)] = t / (4.0 * np.pi)
                    JJ_list[(R, iatom, jatom)] = tmp / (4.0 * np.pi)
                    Rij_done.add((R, iatom, jatom))
                    if (Rm, jatom, iatom) not in Rij_done:
                        Jorb_list[(Rm, jatom, iatom)] = t.T / (4.0 * np.pi)
                        JJ_list[(Rm, jatom, iatom)] = tmp / (4.0 * np.pi)
                        Rij_done.add((Rm, jatom, iatom))
        return Jorb_list, JJ_list

    def A_to_Jtensor(self):
        for key, val in self.JJ.items():
            # key:(R, iatom, jatom)
            R, iatom, jatom = key
            ispin = self.ispin(iatom)
            jspin = self.ispin(jatom)
            keyspin = (R, ispin, jspin)
            is_nonself = not (R == (0, 0, 0) and iatom == jatom)
            Jij = np.imag(val) / np.sign(np.dot(self.spinat[iatom], self.spinat[jatom]))
            Jorbij = np.imag(self.Jorb[key]) / np.sign(
                np.dot(self.spinat[iatom], self.spinat[jatom])
            )
            if is_nonself:
                self.exchange_Jdict[keyspin] = Jij
                self.Jiso_orb[keyspin] = self.simplify_orbital_contributions(
                    Jorbij, iatom, jatom
                )

    def get_rho_e(self, rho_up, rho_dn):
        # self.rho_up_list.append(-1.0 / np.pi * np.imag(rho_up[(0,0,0)]))
        # self.rho_dn_list.append(-1.0 / np.pi * np.imag(rho_dn[(0,0,0)]))
        rup = -1.0 / np.pi * rho_up[(0, 0, 0)]
        rdn = -1.0 / np.pi * rho_dn[(0, 0, 0)]
        return rup, rdn

    def get_rho_atom(self):
        """
        charges and spins from density matrices
        """
        self.charges = np.zeros(len(self.atoms), dtype=float)
        self.spinat = np.zeros((len(self.atoms), 3), dtype=float)
        for iatom in self.orb_dict:
            iorb = self.iorb(iatom)
            tup = np.real(np.trace(self.rho_up[np.ix_(iorb, iorb)]))
            tdn = np.real(np.trace(self.rho_dn[np.ix_(iorb, iorb)]))
            self.charges[iatom] = tup + tdn
            self.spinat[iatom, 2] = tup - tdn

    def finalize(self):
        self.Gup.clean_cache()
        self.Gdn.clean_cache()
        # path = 'TB2J_results/cache'
        # if os.path.exists(path):
        #    shutil.rmtree(path)

    def integrate(self, method="simpson"):
        # if method == "trapezoidal":
        #    integrate = trapezoidal_nonuniform
        # elif method == "simpson":
        #    integrate = simpson_nonuniform
        for R, ijpairs in self.R_ijatom_dict.items():
            for iatom, jatom in ijpairs:
                # self.Jorb[(R, iatom, jatom)] = integrate(
                #    self.contour.path, self.Jorb_list[(R, iatom, jatom)]
                # )
                # self.JJ[(R, iatom, jatom)] = integrate(
                #    self.contour.path, self.JJ_list[(R, iatom, jatom)]
                # )
                self.Jorb[(R, iatom, jatom)] = self.contour.integrate_values(
                    self.Jorb_list[(R, iatom, jatom)]
                )
                self.JJ[(R, iatom, jatom)] = self.contour.integrate_values(
                    self.JJ_list[(R, iatom, jatom)]
                )

    def get_quantities_per_e(self, e):
        GR_up = self.Gup.get_GR(self.short_Rlist, energy=e, get_rho=False)
        GR_dn = self.Gdn.get_GR(self.short_Rlist, energy=e, get_rho=False)
        Jorb_list, JJ_list = self.get_all_A(GR_up, GR_dn)
        return dict(Jorb_list=Jorb_list, JJ_list=JJ_list)

    def calculate_all(self):
        """
        The top level.
        """
        print("Green's function Calculation started.")

        npole = len(self.contour.path)
        if self.nproc == 1:
            results = map(
                self.get_quantities_per_e, tqdm(self.contour.path, total=npole)
            )
        else:
            # pool = ProcessPool(nodes=self.nproc)
            # results = pool.map(self.get_AijR_rhoR ,self.contour.path)
            results = p_map(
                self.get_quantities_per_e, self.contour.path, num_cpus=self.nproc
            )
        for i, result in enumerate(results):
            Jorb_list = result["Jorb_list"]
            JJ_list = result["JJ_list"]
            for iR, R in enumerate(self.R_ijatom_dict):
                for iatom, jatom in self.R_ijatom_dict[R]:
                    key = (R, iatom, jatom)
                    self.Jorb_list[key].append(Jorb_list[key])
                    self.JJ_list[key].append(JJ_list[key])
        self.integrate()
        self.get_rho_atom()
        self.A_to_Jtensor()

    def write_output(self, path="TB2J_results"):
        self._prepare_index_spin()
        output = SpinIO(
            atoms=self.atoms,
            charges=self.charges,
            spinat=self.spinat,
            index_spin=self.index_spin,
            colinear=True,
            orbital_names=self.orbital_names,
            distance_dict=self.distance_dict,
            exchange_Jdict=self.exchange_Jdict,
            Jiso_orb=self.Jiso_orb,
            dmi_ddict=None,
            NJT_Jdict=None,
            NJT_ddict=None,
            Jani_dict=None,
            biquadratic_Jdict=None,
            description=self.description,
        )
        output.write_all(path=path)



import os
import pickle
from collections import defaultdict

import numpy as np
from tqdm import tqdm

from TB2J.contour import Contour
from TB2J.exchange_params import ExchangeParams
from TB2J.external import p_map
from TB2J.green import TBGreen
from TB2J.io_exchange import SpinIO
from TB2J.mycfr import CFR
from TB2J.orbmap import map_orbs_matrix
from TB2J.pauli import pauli_block_all, pauli_block_sigma_norm
from TB2J.utils import (
    kmesh_to_R,
    symbol_number,
)


class Exchange(ExchangeParams):
    def __init__(self, tbmodels, atoms, **params):
        self.atoms = atoms
        super().__init__(**params)
        self._prepare_kmesh(self._kmesh)
        self._prepare_Rlist()
        self.set_tbmodels(tbmodels)
        self._adjust_emin()
        # self._prepare_elist(method="CFR")
        self._prepare_elist(method="legendre")
        self._prepare_basis()
        self._prepare_orb_dict()
        self._prepare_distance()

        # whether to calculate J and DMI with NJt method.
        # self._prepare_NijR()
        self._is_collinear = True
        self.has_elistc = False
        self._clean_tbmodels()

    def _prepare_Jorb_file(self):
        os.makedirs(self.output_path, exist_ok=True)
        self.orbpath = os.path.join(self.output_path, "OrbResolve")
        os.makedirs(self.orbpath, exist_ok=True)

    def _adjust_emin(self):
        self.emin = self.G.find_energy_ingap(rbound=self.efermi - 15.0) - self.efermi
        # self.emin = self.G.find_energy_ingap(rbound=self.efermi - 15.0) - self.efermi
        # self.emin = -42.0
        print(f"A gap is found at {self.emin}, set emin to it.")

    def set_tbmodels(self, tbmodels):
        pass

    def _clean_tbmodels(self):
        # del self.tbmodel
        # del self.G.tbmodel
        pass

    def _prepare_kmesh(self, kmesh, ibz=False):
        for k in kmesh:
            self.kmesh = list(map(lambda x: x // 2 * 2 + 1, kmesh))

    def _prepare_elist(self, method="CFR"):
        """
        prepare list of energy for integration.
        The path has three segments:
         emin --1-> emin + 1j*height --2-> emax+1j*height --3-> emax
        """
        # if method.lower() == "rectangle":
        #    self.contour.build_path_rectangle(
        #        height=self.height, nz1=self.nz1, nz2=self.nz2, nz3=self.nz3
        #    )
        if method.lower() == "semicircle":
            self.contour = Contour(self.emin, self.emax)
            self.contour.build_path_semicircle(npoints=self.nz, endpoint=True)
        elif method.lower() == "legendre":
            self.contour = Contour(self.emin, self.emax)
            self.contour.build_path_legendre(npoints=self.nz, endpoint=True)
        elif method.lower() == "cfr":
            self.contour = CFR(nz=self.nz, T=600)
        else:
            raise ValueError(f"The path cannot be of type {method}.")

    def _prepare_Rlist(self):
        """
        prepare R list for J(i, j, R)
        [-Rx, Rx] * [-Ry, Ry] * [-Rz, Rz]
        """
        self.Rlist = kmesh_to_R(self.kmesh)

    def _prepare_basis(self):
        if self.basis is None:
            pass

    def _prepare_orb_dict(self):
        """
        Generate orbital and magnetic atom mappings needed for exchange calculations.

        Creates:
        - self.orb_dict: Maps atom indices to their orbital indices
        - self.labels: Maps atom indices to their orbital labels
        - self.ind_mag_atoms: List of indices of magnetic atoms
        - self._spin_dict: Maps atom indices to spin indices
        - self._atom_dict: Maps spin indices back to atom indices
        """
        self._create_orbital_mappings()
        self._identify_magnetic_atoms()
        self._validate_orbital_assignments()
        self._create_spin_mappings()
        self._prepare_orb_mmat()

    def _create_orbital_mappings(self):
        """Create mappings between atoms and their orbitals."""
        self.orb_dict = {}  # {atom_index: [orbital_indices]}
        self.labels = {}  # {atom_index: [orbital_labels]}
        atom_symbols = symbol_number(self.atoms)

        for orb_idx, base in enumerate(self.basis):
            if orb_idx in self.exclude_orbs:
                continue

            # Extract atom and orbital info
            if isinstance(base, str):
                atom_sym, orb_sym = base.split("|")[:2]
                atom_idx = atom_symbols[atom_sym]
            else:
                try:
                    atom_sym, orb_sym = base[:2]
                    atom_idx = atom_symbols[atom_sym]
                except Exception:
                    atom_idx = base.iatom
                    atom_sym = base.iatom
                    orb_sym = base.sym

            # Update orbital mappings
            if atom_idx not in self.orb_dict:
                self.orb_dict[atom_idx] = [orb_idx]
                self.labels[atom_idx] = [orb_sym]
            else:
                self.orb_dict[atom_idx].append(orb_idx)
                self.labels[atom_idx].append(orb_sym)

    def _identify_magnetic_atoms(self):
        """Identify which atoms are magnetic based on elements/tags."""
        if self.index_magnetic_atoms is not None:
            self.ind_mag_atoms = self.index_magnetic_atoms
        else:
            self.ind_mag_atoms = []
            symbols = self.atoms.get_chemical_symbols()
            tags = self.atoms.get_tags()

            for atom_idx, (sym, tag) in enumerate(zip(symbols, tags)):
                if (
                    sym in self.magnetic_elements
                    or f"{sym}{tag}" in self.magnetic_elements
                ):
                    self.ind_mag_atoms.append(atom_idx)

    def _validate_orbital_assignments(self):
        """Validate that magnetic atoms have proper orbital assignments."""
        # Check all magnetic atoms have orbitals
        for atom_idx in self.ind_mag_atoms:
            if atom_idx not in self.orb_dict:
                raise ValueError(
                    f"Atom {atom_idx} is magnetic but has no orbitals assigned. "
                    "Check Wannier function definitions."
                )

        # For non-collinear case, check spin-orbital pairing
        if not self._is_collinear:
            for atom_idx, orbitals in self.orb_dict.items():
                if len(orbitals) % 2 != 0:
                    raise ValueError(
                        f"Atom {atom_idx} has {len(orbitals)} spin-orbitals "
                        "(should be even). Check Wannier function localization."
                    )

    def _create_spin_mappings(self):
        """Create mappings between atom indices and spin indices."""
        self._spin_dict = {}  # {atom_index: spin_index}
        self._atom_dict = {}  # {spin_index: atom_index}

        for spin_idx, atom_idx in enumerate(self.ind_mag_atoms):
            self._spin_dict[atom_idx] = spin_idx
            self._atom_dict[spin_idx] = atom_idx

    def _prepare_orb_mmat(self):
        self.mmats = {}
        self.orbital_names = {}
        self.norb_reduced = {}
        if self.backend_name.upper() in ["SIESTA", "ABACUS", "LCAOHAMILTONIAN"]:
            # print(f"magntic_elements: {self.magnetic_elements}")
            # print(f"include_orbs: {self.include_orbs}")
            syms = self.atoms.get_chemical_symbols()
            for iatom, orbs in self.labels.items():
                if (self.include_orbs is not None) and syms[iatom] in self.include_orbs:
                    mmat, reduced_orbs = map_orbs_matrix(
                        orbs,
                        spinor=not (self._is_collinear),
                        include_only=self.include_orbs[syms[iatom]],
                    )
                else:
                    # print(f"orbs: {orbs}")
                    mmat, reduced_orbs = map_orbs_matrix(
                        orbs, spinor=not (self._is_collinear), include_only=None
                    )

                self.mmats[iatom] = mmat
                self.orbital_names[iatom] = reduced_orbs
                # Note that for siesta, spin up and spin down has same orb name.
                # Therefor there is no nedd to /2
                self.norb_reduced[iatom] = len(reduced_orbs)
        else:
            self.orbital_names = self.labels
            for iatom, orbs in self.labels.items():
                # Note that for siesta, spin up and spin down has same orb name.
                # thus //2
                self.norb_reduced[iatom] = len(orbs) // 2

    def ispin(self, iatom):
        return self._spin_dict[iatom]

    def iatom(self, ispin):
        return self._atom_dict[ispin]

    def _prepare_distance(self):
        """
        prepare the distance between atoms.
        """
        self.distance_dict = {}
        self.short_Rlist = []
        self.R_ijatom_dict = defaultdict(lambda: [])
        ind_matoms = self.ind_mag_atoms
        for ispin, iatom in enumerate(ind_matoms):
            for jspin, jatom in enumerate(ind_matoms):
                for R in self.Rlist:
                    pos_i = self.atoms.get_positions()[iatom]
                    pos_jR = self.atoms.get_positions()[jatom] + np.dot(
                        R, self.atoms.get_cell()
                    )
                    vec = pos_jR - pos_i
                    distance = np.sqrt(np.sum(vec**2))
                    if self.Rcut is None or distance < self.Rcut:
                        self.distance_dict[(tuple(R), ispin, jspin)] = (vec, distance)
                        self.R_ijatom_dict[tuple(R)].append((iatom, jatom))
        self.short_Rlist = list(self.R_ijatom_dict.keys())

    def iorb(self, iatom):
        """
        given an index of atom, return all the orbital indices of it.
        :param iatom:  index of atom.
        """
        return np.array(self.orb_dict[iatom], dtype=int)

    def simplify_orbital_contributions(self, Jorbij, iatom, jatom):
        """
        sum up the contribution of all the orbitals with same (n,l,m)
        """
        if self.backend_name.upper() in ["SIESTA", "ABACUS", "LCAOHAMILTONIAN"]:
            mmat_i = self.mmats[iatom]
            mmat_j = self.mmats[jatom]
            Jorbij = mmat_i.T @ Jorbij @ mmat_j
        return Jorbij

    def calculate_all(self):
        raise NotImplementedError(
            "calculate all is not implemented in a abstract class. "
        )

    def write_output(self):
        raise NotImplementedError(
            "write_output is not implemented in a abstract class. "
        )


class ExchangeNCL(Exchange):
    """
    Non-collinear exchange
    """

    def set_tbmodels(self, tbmodels):
        """
        tbmodels should be in spinor form.
        The basis should be orb1_up, orb2_up,...orbn_up, orb1_dn, orb2_dn....
        """
        self.tbmodel = tbmodels
        self.backend_name = self.tbmodel.name
        self.G = TBGreen(
            tbmodel=self.tbmodel,
            kmesh=self.kmesh,
            ibz=self.ibz,
            gamma=True,
            efermi=self.efermi,
            use_cache=self._use_cache,
            nproc=self.nproc,
        )
        if self.efermi is None:
            self.efermi = self.G.efermi
        self.norb = self.G.norb
        self.nbasis = self.G.nbasis
        # self.rho = np.zeros((self.nbasis, self.nbasis), dtype=complex)
        self.rho = self.G.get_density_matrix()
        self.A_ijR_list = defaultdict(lambda: [])
        self.A_ijR = defaultdict(lambda: np.zeros((4, 4), dtype=complex))
        self.A_ijR_orb = dict()
        # self.HR0 = self.tbmodel.get_H0()
        if hasattr(self.tbmodel, "get_H0"):
            self.HR0 = self.tbmodel.get_H0()
        else:
            self.HR0 = self.G.H0
        self._is_collinear = False
        self.Pdict = {}
        if self.write_density_matrix:
            self.G.write_rho_R()

    def get_MAE(self, thetas, phis):
        """
        Calculate the magnetic anisotropy energy.
        """
        pass

    def _prepare_NijR(self):
        self.N = {}
        for R in self.Rlist:
            self.N[R] = np.zeros((self.nbasis, self.nbasis), dtype=complex)

    def _prepare_Patom(self):
        for iatom in self.ind_mag_atoms:
            self.Pdict[iatom] = pauli_block_sigma_norm(self.get_H_atom(iatom))

    def get_H_atom(self, iatom):
        orbs = self.iorb(iatom)
        # return self.HR0[self.orb_slice[iatom], self.orb_slice[iatom]]
        return self.HR0[np.ix_(orbs, orbs)]

    def get_P_iatom(self, iatom):
        """Calculate the norm of the Hamiltonian vector.
        For each atom, the local hamiltonian of each orbital H(2*2 matrix)
        can be written as H0* I + H1* sigma1 + H2*sigma2 + H3 *sigma3
        where sigma is a Pauli matrix. return the norm of (H1, H2, H3) vector
        :param iatom: index of atom
        :returns: a matrix of norms P. P[i, j] is for orbital i and j.
        :rtype: complex matrix of shape norb_i * norb_i.
        """
        if self.Pdict == {}:
            self._prepare_Patom()
        return self.Pdict[iatom]

    def GR_atom(self, GR, iatom, jatom):
        """Given a green's function matrix, return the [iatom, jatom] component.

        :param GR:  Green's function matrix
        :param iatom:  index of atom i
        :param jatom:  index of atom j
        :returns:  G_ij
        :rtype:  complex matrix.

        """
        orbi = self.iorb(iatom)
        orbj = self.iorb(jatom)
        return GR[np.ix_(orbi, orbj)]
        # return GR[self.orb_slice[iatom], self.orb_slice[jatom]]

    def get_A_ijR(self, G, R, iatom, jatom):
        """calculate A from G for a energy slice (de).
        It take the
        .. math::
           A^{uv} = p T^u p T^v dE / pi

        where u, v are I, x, y, z (index 0, 1,2,3). p(i) = self.get_P_iatom(iatom)
        T^u(ijR)  (u=0,1,2,3) = pauli_block_all(G)

        :param G: Green's function for all R, i, j.
        :param iatom: i
        :param jatom: j
        :param de:  energy step. used for integeration
        :returns: a matrix of A_ij(u, v), where u, v =(0)0, x(1), y(2), z(3)
        :rtype:  4*4 matrix
        """
        GR = G[R]
        Gij = self.GR_atom(GR, iatom, jatom)
        Gij_Ixyz = pauli_block_all(Gij)

        # G(j, i, -R)
        Rm = tuple(-x for x in R)
        GRm = G[Rm]
        Gji = self.GR_atom(GRm, jatom, iatom)
        Gji_Ixyz = pauli_block_all(Gji)

        ni = self.norb_reduced[iatom]
        nj = self.norb_reduced[jatom]

        tmp = np.zeros((4, 4), dtype=complex)

        if self.orb_decomposition:
            torb = np.zeros((4, 4, ni, nj), dtype=complex)
            # a, b in (0,x,y,z)
            for a in range(4):
                piGij = self.get_P_iatom(iatom) @ Gij_Ixyz[a]
                for b in range(4):
                    pjGji = self.get_P_iatom(jatom) @ Gji_Ixyz[b]
                    torb[a, b] = self.simplify_orbital_contributions(
                        np.einsum("ij, ji -> ij", piGij, pjGji) / np.pi, iatom, jatom
                    )
                    tmp[a, b] = np.sum(torb[a, b])

        else:
            for a in range(4):
                pGp = self.get_P_iatom(iatom) @ Gij_Ixyz[a] @ self.get_P_iatom(jatom)
                for b in range(4):
                    AijRab = pGp @ Gji_Ixyz[b]
                    tmp[a, b] = np.trace(AijRab) / np.pi
            torb = None
        return tmp, torb

    def get_all_A(self, G):
        """
        Calculate all A matrix elements
        Loop over all magnetic atoms.
        :param G: Green's function.
        :param de: energy step.
        """
        A_ijR_list = {}
        Aorb_ijR_list = {}
        for iR, R in enumerate(self.R_ijatom_dict):
            for iatom, jatom in self.R_ijatom_dict[R]:
                A, A_orb = self.get_A_ijR(G, R, iatom, jatom)
                A_ijR_list[(R, iatom, jatom)] = A
                Aorb_ijR_list[(R, iatom, jatom)] = A_orb
        return A_ijR_list, Aorb_ijR_list

    def A_to_Jtensor_orb(self):
        """
        convert the orbital composition of A into J, DMI, Jani
        """
        self.Jiso_orb = {}
        self.Jani_orb = {}
        self.DMI_orb = {}

        if self.orb_decomposition:
            for key, val in self.A_ijR_orb.items():
                R, iatom, jatom = key
                # Rm = tuple(-x for x in R)
                # valm = self.A_ijR_orb[(Rm, jatom, iatom)]
                ni = self.norb_reduced[iatom]
                nj = self.norb_reduced[jatom]

                is_nonself = not (R == (0, 0, 0) and iatom == jatom)
                ispin = self.ispin(iatom)
                jspin = self.ispin(jatom)
                keyspin = (R, ispin, jspin)

                # isotropic J
                Jiso = np.imag(val[0, 0] - val[1, 1] - val[2, 2] - val[3, 3])

                # off-diagonal anisotropic exchange
                Ja = np.zeros((3, 3, ni, nj), dtype=float)
                for i in range(3):
                    for j in range(3):
                        Ja[i, j] = np.imag(val[i + 1, j + 1] + val[j + 1, i + 1])
                # DMI

                Dtmp = np.zeros((3, ni, nj), dtype=float)
                for i in range(3):
                    Dtmp[i] = np.real(val[0, i + 1] - val[i + 1, 0])

                if is_nonself:
                    self.Jiso_orb[keyspin] = Jiso
                    self.Jani_orb[keyspin] = Ja
                    self.DMI_orb[keyspin] = Dtmp

    def A_to_Jtensor(self):
        """
        Calculate J tensors from A.
        If we assume the exchange can be written as a bilinear tensor form,
        J_{isotropic} = Tr Im (A^{00} - A^{xx} - A^{yy} - A^{zz})
        J_{anisotropic}_uv = Tr Im (2A)
        DMI =  Tr Re (A^{0z} - A^{z0} )
        """
        self.Jani = {}
        self.DMI = {}

        self.Jprime = {}
        self.B = {}
        self.exchange_Jdict = {}
        self.Jiso_orb = {}

        self.debug_dict = {"DMI2": {}}

        for key, val in self.A_ijR.items():
            # key:(R, iatom, jatom)
            R, iatom, jatom = key

            Rm = tuple(-x for x in R)
            valm = self.A_ijR[(Rm, jatom, iatom)]

            ispin = self.ispin(iatom)
            jspin = self.ispin(jatom)
            keyspin = (R, ispin, jspin)

            is_nonself = not (R == (0, 0, 0) and iatom == jatom)
            Jiso = 0.0
            Ja = np.zeros((3, 3), dtype=float)
            Dtmp = np.zeros(3, dtype=float)
            Dtmp2 = np.zeros(3, dtype=float)
            # Heisenberg like J.
            Jiso = np.imag(val[0, 0] - val[1, 1] - val[2, 2] - val[3, 3])

            if is_nonself:
                self.exchange_Jdict[keyspin] = Jiso

            # off-diagonal anisotropic exchange
            for i in range(3):
                for j in range(3):
                    Ja[i, j] = np.imag(val[i + 1, j + 1] + valm[i + 1, j + 1])
            if is_nonself:
                self.Jani[keyspin] = Ja

            # DMI
            for i in range(3):
                Dtmp[i] = np.real(val[0, i + 1] - val[i + 1, 0])

            # Dx = Jyz-Jzy
            # Dy = Jzx-Jxz
            # Dz = Jxy-Jyx
            Dtmp2[0] = np.imag(val[2, 3] - val[3, 2])
            Dtmp2[1] = np.imag(val[3, 1] - val[1, 3])
            Dtmp2[2] = np.imag(val[1, 2] - val[2, 1])
            if is_nonself:
                self.DMI[keyspin] = Dtmp
                self.debug_dict["DMI2"][keyspin] = Dtmp2

            # isotropic exchange into bilinear and biqudratic parts:
            # Jprime SiSj and B (SiSj)^2
            if is_nonself:
                Si = self.spinat[iatom]
                Sj = self.spinat[jatom]
                Jprime = np.imag(val[0, 0] - val[3, 3]) - 2 * np.sign(
                    np.dot(Si, Sj)
                ) * np.imag(val[3, 3])
                # Jprime = np.imag(val[0, 0] - 3*val[3, 3])
                B = np.imag(val[3, 3])
                self.B[keyspin] = Jprime, B

    # def get_N_e(self, GR, de):
    #    """
    #    calcualte density matrix for all R,i, j
    #    """
    #    self.N = defaultdict(lambda: 0.0)
    #    for R, G in GR.items():
    #        self.N[R] += -1.0 / np.pi * np.imag(G * de)

    # def get_rho_e(self, rhoR):
    #    """add component to density matrix from a green's function
    #    :param GR: Green's funciton in real space.
    #    """
    #    return -1.0 / np.pi * rhoR[0, 0, 0]

    # def get_total_charges(self):
    #    return np.sum(np.imag(np.diag(self.rho)))

    def get_rho_atom(self):
        """
        calculate charge and spin for each atom.
        """
        rho = {}
        self.charges = np.zeros(len(self.atoms), dtype=float)
        self.spinat = np.zeros((len(self.atoms), 3), dtype=float)
        for iatom in self.orb_dict:
            iorb = self.iorb(iatom)
            tmp = self.rho[np.ix_(iorb, iorb)]
            # *2 because there is a 1/2 in the paui_block_all function
            rho[iatom] = np.array([np.trace(x) * 2 for x in pauli_block_all(tmp)]).real
            self.charges[iatom] = rho[iatom][0]
            self.spinat[iatom, :] = rho[iatom][1:]
        self.rho_dict = rho
        return self.rho_dict

    def integrate(self, AijRs, AijRs_orb=None, method="simpson"):
        """
        AijRs: a list of AijR,
        wherer AijR: array of ((nR, n, n, 4,4), dtype=complex)
        """
        # if method == "trapezoidal":
        #    integrate = trapezoidal_nonuniform
        # elif method == "simpson":
        #    integrate = simpson_nonuniform
        #

        # self.rho = integrate(self.contour.path, rhoRs)
        for iR, R in enumerate(self.R_ijatom_dict):
            for iatom, jatom in self.R_ijatom_dict[R]:
                f = AijRs[(R, iatom, jatom)]
                # self.A_ijR[(R, iatom, jatom)] = integrate(self.contour.path, f)
                self.A_ijR[(R, iatom, jatom)] = self.contour.integrate_values(f)

                if self.orb_decomposition:
                    # self.A_ijR_orb[(R, iatom, jatom)] = integrate(
                    #    self.contour.path, AijRs_orb[(R, iatom, jatom)]
                    # )
                    self.contour.integrate_values(AijRs_orb[(R, iatom, jatom)])

    def get_quantities_per_e(self, e):
        Gk_all = self.G.get_Gk_all(e)
        # mae = self.get_mae_kspace(Gk_all)
        mae = None
        # TODO: get the MAE from Gk_all
        GR = self.G.get_GR(self.short_Rlist, energy=e, get_rho=False, Gk_all=Gk_all)
        # TODO: define the quantities for one energy.
        AijR, AijR_orb = self.get_all_A(GR)
        return dict(AijR=AijR, AijR_orb=AijR_orb, mae=mae)

    def save_AijR(self, AijRs, fname):
        result = dict(path=self.contour.path, AijRs=AijRs)
        with open(fname, "wb") as myfile:
            pickle.dump(result, myfile)

    def validate(self):
        """
        Do some sanity check before proceding.
        """
        pass

    def calculate_all(self):
        """
        The top level.
        """
        print("Green's function Calculation started.")

        AijRs = {}
        AijRs_orb = {}

        self.validate()

        npole = len(self.contour.path)
        if self.nproc > 1:
            results = p_map(
                self.get_quantities_per_e, self.contour.path, num_cpus=self.nproc
            )
        else:
            results = map(
                self.get_quantities_per_e, tqdm(self.contour.path, total=npole)
            )

        for i, result in enumerate(results):
            for iR, R in enumerate(self.R_ijatom_dict):
                for iatom, jatom in self.R_ijatom_dict[R]:
                    if (R, iatom, jatom) in AijRs:
                        AijRs[(R, iatom, jatom)].append(result["AijR"][R, iatom, jatom])
                        if self.orb_decomposition:
                            AijRs_orb[(R, iatom, jatom)].append(
                                result["AijR_orb"][R, iatom, jatom]
                            )

                    else:
                        AijRs[(R, iatom, jatom)] = []
                        AijRs[(R, iatom, jatom)].append(result["AijR"][R, iatom, jatom])
                        if self.orb_decomposition:
                            AijRs_orb[(R, iatom, jatom)] = []
                            AijRs_orb[(R, iatom, jatom)].append(
                                result["AijR_orb"][R, iatom, jatom]
                            )

        # self.save_AijRs(AijRs)
        self.integrate(AijRs, AijRs_orb)
        self.get_rho_atom()
        self.A_to_Jtensor()
        self.A_to_Jtensor_orb()

    def _prepare_index_spin(self):
        # index_spin: index in spin hamiltonian of atom. starts from 1. -1 means not considered.
        ind_matoms = []
        self.index_spin = []
        ispin = 0
        symbols = self.atoms.get_chemical_symbols()
        tags = self.atoms.get_tags()
        for i, (sym, tag) in enumerate(zip(symbols, tags)):
            if sym in self.magnetic_elements or f"{sym}{tag}" in self.magnetic_elements:
                ind_matoms.append(i)
                self.index_spin.append(ispin)
                ispin += 1
            else:
                self.index_spin.append(-1)

    def write_output(self, path="TB2J_results"):
        self._prepare_index_spin()
        output = SpinIO(
            atoms=self.atoms,
            charges=self.charges,
            spinat=self.spinat,
            index_spin=self.index_spin,
            colinear=False,
            orbital_names=self.orbital_names,
            distance_dict=self.distance_dict,
            exchange_Jdict=self.exchange_Jdict,
            Jiso_orb=self.Jiso_orb,
            dmi_ddict=self.DMI,
            Jani_dict=self.Jani,
            DMI_orb=self.DMI_orb,
            Jani_orb=self.Jani_orb,
            biquadratic_Jdict=self.B,
            debug_dict=self.debug_dict,
            description=self.description,
        )
        output.write_all(path=path)
        # with open("TB2J_results/J_orb.pickle", 'wb') as myfile:
        #    pickle.dump({'Jiso_orb': self.Jiso_orb,
        #                 'DMI_orb': self.DMI_orb, 'Jani_orb': self.Jani_orb}, myfile)

    def finalize(self):
        self.G.clean_cache()

    def run(self, path="TB2J_results"):
        self.calculate_all()
        self.write_output(path=path)
        self.finalize()


class ExchangeCL(ExchangeNCL):
    def set_tbmodels(self, tbmodel):
        """
        only difference is a colinear tag.
        """
        super().set_tbmodels(tbmodel)
        self._is_collinear = True

    def write_output(self, path="TB2J_results"):
        self._prepare_index_spin()
        output = SpinIO(
            atoms=self.atoms,
            charges=self.charges,
            spinat=self.spinat,
            index_spin=self.index_spin,
            orbital_names=self.orbital_names,
            colinear=True,
            distance_dict=self.distance_dict,
            exchange_Jdict=self.exchange_Jdict,
            dmi_ddict=None,
            biquadratic_Jdict=self.B,
            description=self.description,
        )
        output.write_all(path=path)



import sys

import numpy as np

MAX_EXP_ARGUMENT = np.log(sys.float_info.max / 100000)
print(MAX_EXP_ARGUMENT)


def fermi(e, mu, width=0.01):
    """
    the fermi function.
     .. math::
        f=\\frac{1}{\exp((e-\mu)/width)+1}

    :param e,mu,width: e,\mu,width
    """
    x = (e - mu) / width
    # disable overflow warning
    # with warnings.catch_warnings():
    #    warnings.simplefilter("ignore")
    #    ret = np.where(x < MAX_EXP_ARGUMENT, 1 / (1.0 + np.exp(x)), 0.0)

    ret = np.zeros_like(x, dtype=float)
    for i, xi in enumerate(x):
        if xi < 700:
            ret[i] = 1 / (1.0 + np.exp(xi))
    return ret



import copy
import os
import pickle
import sys
import tempfile
from collections import defaultdict
from shutil import rmtree

import numpy as np
from HamiltonIO.model.occupations import GaussOccupations
from HamiltonIO.model.occupations import myfermi as fermi
from pathos.multiprocessing import ProcessPool

from TB2J.kpoints import ir_kpts, monkhorst_pack

# from TB2J.mathutils.fermi import fermi

MAX_EXP_ARGUMENT = np.log(sys.float_info.max)


def eigen_to_G2(H, S, efermi, energy):
    """calculate green's function from eigenvalue/eigenvector for energy(e-ef): G(e-ef).
    :param H: Hamiltonian matrix in eigenbasis
    :param S: Overlap matrix in eigenbasis
    :param efermi: fermi energy
    :param energy: energy level e - efermi
    """
    # G = ((E+Ef) S - H)^-1
    return np.linalg.inv((energy + efermi) * S - H)


def eigen_to_G(evals, evecs, efermi, energy):
    """calculate green's function from eigenvalue/eigenvector for energy(e-ef): G(e-ef).
    :param evals:  eigen values
    :param evecs:  eigen vectors
    :param efermi: fermi energy
    :param energy: energy
    :returns: Green's function G,
    :rtype:  Matrix with same shape of the Hamiltonian (and eigenvector)
    """
    # return (
    #    np.einsum("ij, j-> ij", evecs, 1.0 / (-evals + (energy + efermi)))
    #    @ evecs.conj().T
    # )
    return np.einsum(
        "ib, b, jb-> ij",
        evecs,
        1.0 / (-evals + (energy + efermi)),
        evecs.conj(),
        optimize=True,
    )


def find_energy_ingap(evals, rbound, gap=2.0):
    """
    find a energy inside a gap below rbound (right bound),
    return the energy gap top - 0.5.
    """
    m0 = np.sort(evals.flatten())
    m = m0[m0 < rbound]
    ind = np.where(np.diff(m) > gap)[0]
    if len(ind) == 0:
        return m0[0] - 0.5
    else:
        return m[ind[-1] + 1] - 0.5


class TBGreen:
    def __init__(
        self,
        tbmodel,
        kmesh=None,  # [ikpt, 3]
        ibz=False,  # if True, will interpolate the Green's function at Ir-kpoints
        efermi=None,  # efermi
        gamma=False,
        kpts=None,
        kweights=None,
        k_sym=False,
        use_cache=False,
        cache_path=None,
        nproc=1,
    ):
        """
        :param tbmodel: A tight binding model
        :param kmesh: size of monkhorst pack. e.g [6,6,6]
        :param efermi: fermi energy.
        """
        self.tbmodel = tbmodel
        self.is_orthogonal = tbmodel.is_orthogonal
        self.R2kfactor = tbmodel.R2kfactor
        self.k2Rfactor = -tbmodel.R2kfactor
        self.efermi = efermi
        self._use_cache = use_cache
        self.cache_path = cache_path
        if use_cache:
            self._prepare_cache()
        self.prepare_kpts(
            kmesh=kmesh,
            ibz=ibz,
            gamma=gamma,
            kpts=kpts,
            kweights=kweights,
            tbmodel=tbmodel,
        )

        self.norb = tbmodel.norb
        self.nbasis = tbmodel.nbasis
        self.k_sym = k_sym
        self.nproc = nproc
        self._prepare_eigen()

    def prepare_kpts(
        self, kmesh=None, gamma=True, ibz=False, kpts=None, kweights=None, tbmodel=None
    ):
        """
        Prepare the k-points for the calculation.

        Parameters:
        - kmesh (tuple): The k-mesh used to generate k-points.
        - gamma (bool): Whether to include the gamma point in the k-points.
        - ibz (bool): Whether to use the irreducible Brillouin zone.
        - kpts (list): List of user-defined k-points.
        - kweights (list): List of weights for each k-point.

        Returns:
        None
        """
        if kpts is not None:
            self.kpts = kpts
            self.nkpts = len(self.kpts)
            self.kweights = kweights
        elif kmesh is not None:
            if ibz:
                self.kpts, self.kweights = ir_kpts(
                    atoms=tbmodel.atoms,
                    mp_grid=kmesh,
                    ir=True,
                    is_time_reversal=False,
                )
                self.nkpts = len(self.kpts)
                print(f"Using IBZ of kmesh of {kmesh}")
                print(f"Number of kpts: {self.nkpts}")
                for kpt, weight in zip(self.kpts, self.kweights):
                    # format the kpt and weight, use 5 digits
                    print(f"{kpt[0]:8.5f} {kpt[1]:8.5f} {kpt[2]:8.5f} {weight:8.5f}")
            else:
                self.kpts = monkhorst_pack(kmesh, gamma_center=gamma)
                self.nkpts = len(self.kpts)
                self.kweights = np.array([1.0 / self.nkpts] * self.nkpts)
        else:
            self.kpts = tbmodel.get_kpts()
            self.nkpts = len(self.kpts)
            self.kweights = np.array([1.0 / self.nkpts] * self.nkpts)

    def _reduce_eigens(self, evals, evecs, emin, emax):
        ts = np.logical_and(evals >= emin, evals < emax)
        ts = np.any(ts, axis=0)
        ts = np.where(ts)[0]
        if len(ts) == 0:
            raise ValueError(
                f"Cannot find any band in the energy range specified by emin {emin} and emax {emax}, which are relative to the Fermi energy. Please check that the Fermi energy, the emin and emax are correct. If you're using Wannier90 output, check the Wannier functions give the right band structure."
            )
        istart, iend = ts[0], ts[-1] + 1
        return evals[:, istart:iend], evecs[:, :, istart:iend]

    def find_energy_ingap(self, rbound, gap=2.0):
        return find_energy_ingap(self.evals, rbound, gap)

    def _prepare_cache(self):
        if self.cache_path is None:
            if "TMPDIR" in os.environ:
                rpath = os.environ["TMPDIR"]
            else:
                rpath = "/dev/shm/TB2J_cache"
        else:
            rpath = self.cache_path
        if not os.path.exists(rpath):
            os.makedirs(rpath)
        self.cache_path = tempfile.mkdtemp(prefix="TB2J", dir=rpath)
        print(f"Writting wavefunctions and Hamiltonian in cache {self.cache_path}")

    def clean_cache(self):
        if (self.cache_path is not None) and os.path.exists(self.cache_path):
            rmtree(self.cache_path)

    def _prepare_eigen(self, solve=True, saveH=False):
        """
        calculate eigen values and vectors for all kpts and save.
        Note that the convention 2 is used here, where the
        phase factor is e^(ik.R), not e^(ik.(R+rj-ri))
        """
        nkpts = len(self.kpts)
        self.evals = np.zeros((nkpts, self.nbasis), dtype=float)
        self.nkpts = nkpts
        self.H0 = np.zeros((self.nbasis, self.nbasis), dtype=complex)
        self.evecs = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
        self.H = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
        if not self.is_orthogonal:
            self.S = np.zeros((nkpts, self.nbasis, self.nbasis), dtype=complex)
        else:
            self.S = None
        if self.nproc == 1:
            results = map(self.tbmodel.HSE_k, self.kpts)
        else:
            executor = ProcessPool(nodes=self.nproc)
            results = executor.map(self.tbmodel.HSE_k, self.kpts, [2] * len(self.kpts))
            executor.close()
            executor.join()
            executor.clear()

        for ik, result in enumerate(results):
            if self.is_orthogonal:
                self.H[ik], _, self.evals[ik], self.evecs[ik] = result
            else:
                self.H[ik], self.S[ik], self.evals[ik], self.evecs[ik] = result
            self.H0 += self.H[ik] / self.nkpts

        if not saveH:
            self.H = None

        # get efermi
        if self.efermi is None:
            print("Calculating Fermi energy from eigenvalues")
            print(f"Number of electrons: {self.tbmodel.nel} ")

            # occ = Occupations(
            #    nel=self.tbmodel.nel, width=0.1, wk=self.kweights, nspin=2
            # )
            # self.efermi = occ.efermi(copy.deepcopy(self.evals))
            # print(f"Fermi energy found: {self.efermi}")

            occ = GaussOccupations(
                nel=self.tbmodel.nel, width=0.1, wk=self.kweights, nspin=2
            )
            self.efermi = occ.efermi(copy.deepcopy(self.evals))
            print(f"Fermi energy found: {self.efermi}")

        # self.evals, self.evecs = self._reduce_eigens(
        #    self.evals, self.evecs, emin=self.efermi - 15.0, emax=self.efermi + 10.1
        # )
        if self._use_cache:
            evecs = self.evecs
            self.evecs_shape = self.evecs.shape
            self.evecs = np.memmap(
                os.path.join(self.cache_path, "evecs.dat"),
                mode="w+",
                shape=self.evecs.shape,
                dtype=complex,
            )
            self.evecs[:, :, :] = evecs[:, :, :]
            del self.evecs

            if self.is_orthogonal:
                self.S = None
            else:
                S = self.S
                self.S = np.memmap(
                    os.path.join(self.cache_path, "S.dat"),
                    mode="w+",
                    shape=(nkpts, self.nbasis, self.nbasis),
                    dtype=complex,
                )
                self.S[:] = S[:]
            if not self.is_orthogonal:
                del self.S

    def get_evecs(self, ik):
        if self._use_cache:
            return np.memmap(
                os.path.join(self.cache_path, "evecs.dat"),
                mode="r",
                shape=self.evecs_shape,
                dtype=complex,
            )[ik]
        else:
            return self.evecs[ik]

    def get_evalue(self, ik):
        return self.evals[ik]

    def get_Hk(self, ik):
        if self._use_cache:
            return np.memmap(
                os.path.join(self.cache_path, "H.dat"),
                mode="r",
                shape=(self.nkpts, self.nbasis, self.nbasis),
                dtype=complex,
            )[ik]
        else:
            return self.H[ik]

    def get_Sk(self, ik):
        if self.is_orthogonal:
            return None
        elif self._use_cache:
            return np.memmap(
                os.path.join(self.cache_path, "S.dat"),
                mode="r",
                shape=(self.nkpts, self.nbasis, self.nbasis),
                dtype=complex,
            )[ik]
        else:
            return self.S[ik]

    def get_density_matrix(self):
        rho = np.zeros((self.nbasis, self.nbasis), dtype=complex)
        if self.is_orthogonal:
            for ik, _ in enumerate(self.kpts):
                evecs_k = self.get_evecs(ik)
                # chekc if any of the evecs element is nan
                rho += (
                    (evecs_k * fermi(self.evals[ik], self.efermi, nspin=2))
                    @ evecs_k.T.conj()
                    * self.kweights[ik]
                )
        else:
            for ik, _ in enumerate(self.kpts):
                rho += (
                    (self.get_evecs(ik) * fermi(self.evals[ik], self.efermi, nspin=2))
                    @ self.get_evecs(ik).T.conj()
                    @ self.get_Sk(ik)
                    * self.kweights[ik]
                )
        # check if rho has nan values
        return rho

    def get_rho_R(self, Rlist):
        nR = len(Rlist)
        rho_R = np.zeros((nR, self.nbasis, self.nbasis), dtype=complex)
        for ik, kpt in enumerate(self.kpts):
            evec = self.get_evecs(ik)
            rhok = np.einsum(
                "ib,b, bj-> ij",
                evec,
                fermi(self.evals[ik], self.efermi, nspin=2),
                evec.conj().T,
            )
            for iR, R in enumerate(Rlist):
                rho_R[iR] += rhok * np.exp(self.k2Rfactor * kpt @ R) * self.kweights[ik]
        return rho_R

    def write_rho_R(self, Rlist, fname="rhoR.pickle"):
        rho_R = self.get_rho_R(Rlist)
        with open(fname, "wb") as myfile:
            pickle.dump({"Rlist": Rlist, "rhoR": rho_R}, myfile)

    def get_density(self):
        return np.real(np.diag(self.get_density_matrix()))

    def get_Gk(self, ik, energy, evals=None, evecs=None):
        """Green's function G(k) for one energy
        G(\epsilon)= (\epsilon I- H)^{-1}
        :param ik: indices for kpoint
        :returns: Gk
        :rtype:  a matrix of indices (nbasis, nbasis)
        """
        if evals is None:
            evals = self.get_evalue(ik)
        if evecs is None:
            evecs = self.get_evecs(ik)
        Gk = eigen_to_G(
            evals=evals,
            evecs=evecs,
            efermi=self.efermi,
            energy=energy,
        )

        # A slower version. For test.
        # Gk = np.linalg.inv((energy+self.efermi)*self.S[ik,:,:] - self.H[ik,:,:])
        return Gk

    def get_Gk_all(self, energy):
        """Green's function G(k) for one energy for all kpoints"""
        Gk_all = np.zeros((self.nkpts, self.nbasis, self.nbasis), dtype=complex)
        for ik, _ in enumerate(self.kpts):
            Gk_all[ik] = self.get_Gk(ik, energy)
        return Gk_all

    def get_GR(self, Rpts, energy, get_rho=False, Gk_all=None):
        """calculate real space Green's function for one energy, all R points.
        G(R, epsilon) = G(k, epsilon) exp(-2\pi i R.dot. k)
        :param Rpts: R points
        :param energy:
        :returns:  real space green's function for one energy for a list of R.
        :rtype:  dictionary, the keys are tuple of R, values are matrices of nbasis*nbasis
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0j)
        rhoR = defaultdict(lambda: 0.0j)
        for ik, kpt in enumerate(self.kpts):
            if Gk_all is None:
                Gk = self.get_Gk(ik, energy)
            else:
                Gk = Gk_all[ik]
            if get_rho:
                if self.is_orthogonal:
                    rhok = Gk
                else:
                    rhok = self.get_Sk(ik) @ Gk
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                tmp = Gk * (phase * self.kweights[ik])
                GR[R] += tmp
                # change this if need full rho
                if get_rho and R == (0, 0, 0):
                    rhoR[R] += rhok * (phase * self.kweights[ik])
        if get_rho:
            return GR, rhoR
        else:
            return GR

    def get_GR_and_dGRdx1(self, Rpts, energy, dHdx):
        """
        calculate G(R) and dG(R)/dx.
        dG(R)/dx = \sum_k G(k) (dH(R)/dx) G(k).
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0 + 0.0j)
        dGRdx = defaultdict(lambda: 0.0 + 0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            Gkw = Gk * self.kweights[ik]
            # Gmk = self.get_Gk(self.i_minus_k(kpt), energy)
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                GR[R] += Gkw * (phase * self.kweights[ik])
                dHRdx = dHdx.get_hamR(R)
                dGRdx[R] += Gkw @ dHRdx @ Gk
                # dGRdx[R] += Gk.dot(dHRdx).dot(Gkp)
        return GR, dGRdx

    def get_GR_and_dGRdx(self, Rpts, energy, dHdx):
        """
        calculate G(R) and dG(R)/dx.
        dG(k)/dx =  G(k) (dH(k)/dx) G(k).
        dG(R)/dx = \sum_k dG(k)/dx * e^{-ikR}
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0 + 0.0j)
        dGRdx = defaultdict(lambda: 0.0 + 0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            # Gmk = self.get_Gk(self.i_minus_k(kpt), energy)
            Gkp = Gk * self.kweights[ik]
            dHk = dHdx.gen_ham(tuple(kpt))
            dG = Gk @ dHk @ Gkp
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                GR[R] += Gkp * (phase * self.kweights[ik])
                dGRdx[R] += dG * (phase * self.kweights[ik])
        return GR, dGRdx

    def get_GR_and_dGRdx_and_dGRdx2(self, Rpts, energy, dHdx, dHdx2):
        """
        calculate G(R) and dG(R)/dx.
        dG(k)/dx =  G(k) (dH(k)/dx) G(k).
        dG(R)/dx = \sum_k dG(k)/dx * e^{-ikR}
        """
        Rpts = [tuple(R) for R in Rpts]
        GR = defaultdict(lambda: 0.0 + 0.0j)
        dGRdx = defaultdict(lambda: 0.0 + 0j)
        dGRdx2 = defaultdict(lambda: 0.0 + 0j)
        for ik, kpt in enumerate(self.kpts):
            Gk = self.get_Gk(ik, energy)
            # Gmk = self.get_Gk(self.i_minus_k(kpt), energy)
            Gkp = Gk * self.kweights[ik]
            dHk = dHdx.gen_ham(tuple(kpt))
            dHk2 = dHdx2.gen_ham(tuple(kpt))
            dG = Gk @ dHk @ Gkp
            dG2 = Gk @ dHk2 @ Gkp
            for iR, R in enumerate(Rpts):
                phase = np.exp(self.k2Rfactor * np.dot(R, kpt))
                GR[R] += Gkp * (phase * self.kweights[ik])
                dGRdx[R] += dG * (phase * self.kweights[ik])
                dGRdx2[R] += dG2 * (phase * self.kweights[ik])
        return GR, dGRdx, dGRdx2



import os
import copy
import warnings
import numpy as np
from itertools import combinations_with_replacement, product
from TB2J.io_exchange import SpinIO

u0 = np.zeros(3)
uy = np.array([0.0, 1.0, 0.0])
uz = np.array([0.0, 0.0, 1.0])

def get_Jani_coefficients(a, R=np.eye(3)):

    if len(a) == 1:
        u = a
        v = a
    else:
        u = a[[0, 0, 1]]
        v = a[[0, 1, 1]]
    
    ur = u @ R.T
    vr = v @ R.T
    coefficients = np.hstack([ur*vr, np.roll(ur, -1, axis=-1)*vr + np.roll(vr, -1, axis=-1)*ur])

    return coefficients, u, v

def get_projections(a, b, tol=1e-2):

    projections = np.empty((2, 3))
    if np.linalg.matrix_rank([a, b], tol=tol) == 1:
        if np.linalg.matrix_rank([a, uy], tol=tol) == 1:
            projections[0] = np.cross(a, uz)
        else:
            projections[0] = np.cross(a, uy)
        projections[1] = np.cross(a, projections[0])
        projections /= np.linalg.norm(projections, axis=-1).reshape(-1, 1)
    else:
        projections[0] = np.cross(a, b)
        projections[0] /= np.linalg.norm(projections[0])
        projections[1] = u0

    return projections

class SpinIO_merge(SpinIO):
    def __init__(self, *args, **kwargs):
        super(SpinIO_merge, self).__init__(*args, **kwargs)
        self.projv = None

    def _set_projection_vectors(self):

        norm = np.linalg.norm(self.spinat, axis=-1).reshape(-1, 1)
        spinat = self.spinat / norm
        idx = [self.ind_atoms[i] for i in self.index_spin if i >= 0]
        projv = {}
        for i, j in combinations_with_replacement(range(self.nspin), 2):
            a, b = spinat[idx][[i, j]]
            projv[i, j] = get_projections(a, b)
            projv[j, i] = projv[i, j]

        self.projv = projv

    @classmethod
    def load_pickle(cls, path='TB2J_results', fname='TB2J.pickle'):
        obj = super(SpinIO_merge, cls).load_pickle(path=path, fname=fname)
        obj._set_projection_vectors()

        return obj

def read_pickle(path):
    p1 = os.path.join(path, 'TB2J_results', 'TB2J.pickle')
    p2 = os.path.join(path, 'TB2J.pickle')
    if os.path.exists(p1) and os.path.exists(p2):
        print(f" WARNING!: Both file {p1} and {p2} exist. Use default {p1}.")
    if os.path.exists(p1):
        ret = SpinIO_merge.load_pickle(os.path.join(path, 'TB2J_results'))
    elif os.path.exists(p2):
        ret = SpinIO_merge.load_pickle(path)
    else:
        raise FileNotFoundError(f"Cannot find either file {p1} or {p2}")
    return ret

class Merger():
    def __init__(self, *paths, main_path=None): 
        self.dat = [read_pickle(path) for path in paths]

        if main_path is None:
            self.main_dat = copy.deepcopy(self.dat[-1])
        else:
            self.main_dat = read_pickle(main_path)
            self.dat.append(copy.deepcopy(self.main_dat))

        self._set_projv()
        
    def _set_projv(self):

        cell = self.main_dat.atoms.cell.array
        rotated_cells = np.stack(
            [obj.atoms.cell.array for obj in self.dat], axis=0
        )
        R = np.linalg.solve(cell, rotated_cells)
        indices = range(len(self.dat))

        proju = {}; projv = {}; coeff_matrix = {}; projectors = {};
        for key in self.main_dat.projv.keys():
            vectors = [obj.projv[key] for obj in self.dat]
            coefficients, u, v = zip(*[get_Jani_coefficients(vectors[i], R=R[i]) for i in indices])
            projectors[key] = np.vstack([u[i] @ R[i].T for i in indices])
            coeff_matrix[key] = np.vstack(coefficients)
            proju[key] = np.stack(u)
            projv[key] = np.stack(v)
            if np.linalg.matrix_rank(coeff_matrix[key], tol=1e-2) < 6:
                warnings.warn('''
                    WARNING: The matrix of equations to reconstruct the exchange tensors is
                    close to being singular. This happens when the magnetic moments between
                    different configurations are cloes to being parallel. You need to consider 
                    more rotated spin configurations, otherwise the results might have a large 
                    error.'''
                )
        
        self.proju = proju
        self.projv = projv
        self.coeff_matrix = coeff_matrix
        self.projectors = projectors

    def merge_Jani(self):
        Jani_dict = {}
        proju = self.proju; projv = self.projv; coeff_matrix = self.coeff_matrix;
        for key in self.main_dat.Jani_dict.keys():
            try:
                R, i, j = key
                u = proju[i, j]
                v = projv[i, j]
                Jani = np.stack([sio.Jani_dict[key] for sio in self.dat])
                projections = np.einsum('nmi,nij,nmj->nm', u, Jani, v).flatten()
            except KeyError as err:
                raise KeyError(
                    "Can not find key: %s, Please make sure the three calculations use the same k-mesh and same Rcut."
                    % err)
            newJani = np.linalg.lstsq(coeff_matrix[i, j], projections, rcond=1e-2)[0]
            Jani_dict[key] = np.array([
                [newJani[0], newJani[3], newJani[5]],
                [newJani[3], newJani[1], newJani[4]],
                [newJani[5], newJani[4], newJani[2]]
            ])
        self.main_dat.Jani_dict = Jani_dict

    def merge_Jiso(self):
        Jdict={}
        for key in self.main_dat.exchange_Jdict.keys():
            try:
               J = np.mean([obj.exchange_Jdict[key] for obj in self.dat])
            except KeyError as err:
                raise KeyError(
                        "Can not find key: %s, Please make sure the three calculations use the same k-mesh and same Rcut."
                        % err)
            Jdict[key] = J
        self.main_dat.exchange_Jdict = Jdict
               

    def merge_DMI(self):
        dmi_ddict = {}
        if all(obj.has_dmi for obj in self.dat):
            projectors = self.projectors; proju = self.proju;
            for key in self.main_dat.dmi_ddict.keys():
                try:
                    R, i, j = key
                    u = proju[i, j]
                    DMI = np.stack([sio.dmi_ddict[key] for sio in self.dat])
                    projections = np.einsum('nmi,ni->nm', u, DMI).flatten()
                except KeyError as err:
                    raise KeyError(
                        "Can not find key: %s, Please make sure the three calculations use the same k-mesh and same Rcut."
                        % err)
                newDMI = np.linalg.lstsq(projectors[i, j], projections, rcond=4e-1)[0]
                dmi_ddict[key] = newDMI
            self.main_dat.dmi_ddict = dmi_ddict

def merge(*paths, main_path=None, save=True, write_path='TB2J_results'):
    m = Merger(*paths, main_path=main_path)
    m.merge_Jiso()
    m.merge_DMI()
    m.merge_Jani()
    if save:
        m.main_dat.write_all(path=write_path)
    return m.dat



import numpy as np

from TB2J.io_exchange import SpinIO


class MagnonIO:
    """Handle IO operations for magnon calculations"""

    def __init__(self, exc: SpinIO):
        """
        Initialize MagnonIO with a SpinIO instance

        Parameters
        ----------
        exc : SpinIO
            SpinIO instance containing exchange information
        """
        self.exc = exc

    def get_nspin(self):
        """Get number of spins"""
        return self.exc.get_nspin()

    def get_ind_atoms(self):
        """Get atom indices"""
        return self.exc.ind_atoms

    def get_magmom(self):
        """Get magnetic moments"""
        nspin = self.get_nspin()
        return np.array([self.exc.spinat[self.exc.iatom(i)] for i in range(nspin)])

    def get_rlist(self):
        """Get R-vectors list"""
        return self.exc.Rlist

    def get_jtensor(self, asr=False, iso_only=False):
        """
        Get full J tensor for R-list

        Parameters
        ----------
        asr : bool, optional
            Acoustic sum rule, default False
        iso_only : bool, optional
            Only isotropic interactions, default False
        """
        return self.exc.get_full_Jtensor_for_Rlist(asr=asr, iso_only=iso_only)



import numpy as np

from .io_exchange import ExchangeIO
from .io_exchange.structure import get_attribute_array
from .kpoints import monkhorst_pack
from .mathutils import get_rotation_arrays


def combine_arrays(u, v):
    return np.concatenate(
        [u * v, np.roll(u, -1, axis=-1) * v, np.roll(v, -1, axis=-1) * u], axis=-1
    )


class ExchangeDownfolder(ExchangeIO):
    def __init__(self, **kwargs):
        reference_axes = kwargs.pop("reference_axes", None)
        kwargs["kmesh"] = kwargs.pop("kmesh", [7, 7, 7])
        super().__init__(**kwargs)
        self._old_values = None

        if reference_axes is None:
            reference_axes = np.zeros((6, 3))
            reference_axes[:3] = np.eye(3)
            reference_axes[3:] = np.array(
                [[0.0, 1.0, 1.0], [1.0, 0.0, 1.0], [1.0, 1.0, 0.0]]
            ) / np.sqrt(2)
        self.reference_axes = reference_axes

    @property
    def reference_axes(self):
        return self._axes

    @reference_axes.setter
    def reference_axes(self, values):
        axes = get_attribute_array(values, "reference_axes")
        if axes.ndim != 2 or axes.shape[-1] != 3:
            raise ValueError("The reference axes must be an array of shape (n, 3)")
        self._axes = axes

    @property
    def kpoints(self):
        return self._kpoints

    @ExchangeIO.kmesh.setter
    def kmesh(self, value):
        ExchangeIO.kmesh.fset(self, value)
        self._kpoints = monkhorst_pack(self._kmesh)

    def set_downfolded_magnetic_sites(self, metals):
        # vectors = self.vectors
        old_pairs = self.interacting_pairs
        self._old_magnetic_elements = self.magnetic_elements

        self.magnetic_elements = metals
        indices = [old_pairs.index(pair) for pair in self.interacting_pairs]
        self._old_values = self._exchange_values.copy()
        self.set_vectors(values=self.vectors[indices])

    def _generate_u_matrix(self):
        """
        Constructs the matrix with the coefficients that relate the exchange tensor J_{ij} to the matrix A_{ij}.
        These coefficients only depend on the vectors u_i which depend on the orientations of the magnetic moments.

        """
        i, j = self.i, self.j
        if self.collinear:
            flat_magmoms = self.magmoms[self._index_spin]
            magmoms = np.zeros((flat_magmoms.shape[0], 3))
            magmoms[:, 2] = flat_magmoms
        else:
            magmoms = self.magmoms[self._index_spin]
        magmoms /= np.linalg.norm(magmoms, axis=-1)[:, None]

        U, _ = zip(
            *[get_rotation_arrays(magmoms, u=u[None, :]) for u in self.reference_axes]
        )
        U = np.stack(U).swapaxes(0, 1)

        u1 = combine_arrays(U[i], U[j].conj())
        ur = combine_arrays(U[i], U[j])
        ui = combine_arrays(U[i].conj(), U[j].conj())
        u2 = combine_arrays(U[i].conj(), U[j])
        u = np.concatenate([u1, ur, ui, u2], axis=1)

        self.u_matrix = u

    def _compute_AB_coefficients(self, Hk):
        """
        Computes the coefficients corresponding to A_{ij}(d), B_{ij}(d) from the dynamical matrix h(k). They are
        required to reconstruct the exchange tensor J_{ij}(d).
        These 18 coefficients are stored in an array with the shape (npairs, 18, nvectors)

        """
        i, j = self.i, self.j
        n = i.max() + 1

        ABk = Hk[:, :, [i, i, i + n, i + n], [j, j + n, j, j + n]]
        ABk = np.moveaxis(ABk, [2, 3], [1, 0]).reshape(len(i), 24, -1)

        exp_summand = np.exp(2j * np.pi * self.vectors @ self.kpoints.T)
        AB = np.einsum("nik,ndk->nid", ABk[..., ::-1], exp_summand) / len(self.kpoints)

        self.AB_coefficients = AB

    def compute_exchange_tensor(self, Hk):
        """
        Computes the exchange tensor that best fits the dynamical matrix Hk

        Parameters
        ----------
        Hk : ndarray
            Dynamical matrix corresponding to the points on a k-grid, constructed for different reference axes.
            It has the shape (naxes, nkpoints, 2*natoms, 2*natoms)

        Returns
        -------
        output : ndarray
            An exchange tensor with shape (npairs, ninteractions, 3, 3)
        """
        n = Hk.shape[-1] // 2
        # diag = np.diag_indices(n)
        self.i, self.j = np.triu_indices(n)

        self._generate_u_matrix()
        self._compute_AB_coefficients(Hk)

        ii = np.where(self.i == self.j)
        i0 = np.where((self.vectors == 0).all(axis=-1))
        J = np.stack(
            [
                np.linalg.lstsq(mat, coeffs, rcond=None)[0]
                for mat, coeffs in zip(self.u_matrix, self.AB_coefficients)
            ]
        ).swapaxes(1, 2)
        J = J[:, :, [0, 6, 5, 3, 1, 7, 8, 4, 2]].reshape(len(self.i), -1, 3, 3)
        J *= -1
        J[ii] *= 2
        J[i0] *= 0

        del self.i, self.j, self.u_matrix, self.AB_coefficients

        return J

    def set_exchange_tensor(self, J):
        idig = np.diag_indices(3)
        Jiso = J[:, :, *idig].mean(axis=-1)

        idmi = ([1, 2, 0], [2, 0, 1])
        DMI = (J - J.swapaxes(-1, -2)) / 2

        Jani = (J + J.swapaxes(-1, -2)) / 2
        Jani[:, :, *idig] -= Jiso[:, :, None]

        self._exchange_values[:, :, 3] = Jiso
        if not self.collinear:
            self._exchange_values[:, :, 6:9] = DMI[:, :, *idmi]
            self._exchange_values[:, :, 9:] = Jani.reshape(Jani.shape[:2] + (9,))

    @staticmethod
    def downfold_matrix(matrix, basis):
        eigvals, eigvecs = np.linalg.eigh(matrix)
        A = np.einsum("...ki,...kj->...ij", eigvecs.conj(), basis)
        W, _, Vh = np.linalg.svd(A, full_matrices=False)
        U = np.einsum("...ik,...kj->...ij", W, Vh)
        downfolded_matrix = np.einsum(
            "...ki,...kj->...ij", U.conj(), eigvals[..., None] * U
        )

        return downfolded_matrix

    @staticmethod
    def lowdin_partition(matrix, indices):
        N = matrix.shape[-1] // 2
        null_indices = np.array([i for i in range(N) if i not in indices])
        diag_indices = np.diag_indices(2 * null_indices.size)

        idx = np.concatenate([indices, indices + N])[None, :]
        jdx = np.concatenate([null_indices, null_indices + N])[None, :]

        Hii = matrix[..., idx.T, idx]
        Hij = matrix[..., idx.T, jdx]
        Hji = matrix[..., jdx.T, idx]
        Hjj = matrix[..., jdx.T, jdx]

        eigvals = np.linalg.eigvalsh(matrix)
        Hjj[..., *diag_indices] -= eigvals.min()
        correction = np.einsum("...ij,...jk,...kl->...il", Hij, np.linalg.inv(Hjj), Hji)

        return Hii - correction

    def downfold(self, metals, **params):
        try:
            metals = metals.split()
        except AttributeError:
            try:
                metals = list(metals)
            except (ValueError, TypeError):
                raise TypeError("argument must be a list of element symbols.")

        if any(metal not in self.magnetic_elements for metal in metals):
            wrong_symbols = [
                metal for metal in metals if metal not in self.magnetic_elements
            ]
            raise ValueError(
                f"The metal symbols '{wrong_symbols}' are not magnetic elements."
            )
        else:
            magnetic_sites = [
                symbol for symbol in self.elements if symbol in self.magnetic_elements
            ]
            # nsites = len(magnetic_sites)
            metal_indices = np.array(
                [i for i, element in enumerate(magnetic_sites) if element in metals]
            )

        Hq = np.stack(
            [self.Hq(self.kpoints, u=u[None, :]) for u in self.reference_axes]
        )
        downfolded_Hq = self.lowdin_partition(Hq, metal_indices)

        self.set_downfolded_magnetic_sites(metals)
        J = self.compute_exchange_tensor(downfolded_Hq)
        self.set_exchange_tensor(J.real)

        return J

    def reset(self):
        if self.old_values is not None:
            self.magnetic_elements = self._old_magnetic_elements
            self._exchange_values = self._old_values
            self.old_values = None
            self.old_magnetic_elements = None



import matplotlib.pyplot as plt
from TB2J.spinham.spin_api import SpinModel
from TB2J.io_exchange.io_exchange import SpinIO
import numpy as np
from TB2J import __version__
from TB2J.spinham.qsolver import QSolverASEWrapper
import argparse
import os


def write_eigen(qmesh, gamma=True, path="./", output_fname="EigenJq.txt", **kwargs):
    m = SpinIO.load_pickle(path)
    m.write_Jq(kmesh=qmesh, path=path, gamma=gamma, output_fname=output_fname, **kwargs)


def plot_magnon_band(
    fname="exchange.xml",
    path="./",
    npoints=301,
    show=True,
    kvectors=None,
    knames=None,
    figfname="magnon_band.pdf",
    supercell_matrix=np.eye(3),
    Jq=False,
    kpath_fname="exchange_kpth.txt",
    ax=None,
    **kwargs
):
    m = SpinModel(fname=fname, sc_matrix=None)
    m.set_ham(**kwargs)
    m.plot_magnon_band(
        kvectors=kvectors,
        knames=knames,
        npoints=npoints,
        kpath_fname=kpath_fname,
        Jq=Jq,
        supercell_matrix=supercell_matrix,
        ax=ax,
    )

    plt.savefig(figfname)
    if show:
        plt.show()


def plot_magnon_dos(
    fname="exchange.xml",
    path="./",
    npoints=301,
    window=None,
    smearing_width=0.1,
    kmesh=[9, 9, 9],
    gamma=True,
    show=True,
    figfname="magnon_dos.pdf",
    txt_filename="magnon_dos.txt",
    Jq=False,
    ax=None,
    **kwargs
):
    ffname = os.path.join(path, "exchange.xml")
    if not (os.path.exists(ffname) and os.path.isfile(ffname)):
        ffname = os.path.join(path, "Multibinit", "exchange.xml")
    model = SpinModel(ffname)
    solver = QSolverASEWrapper(model.ham)
    solver.set(kmesh=kmesh, gamma=gamma, Jq=Jq)
    solver.plot_dos(
        npts=npoints,
        window=window,
        output=figfname,
        smearing_width=smearing_width,
        ax=ax,
        show=show,
        dos_filename=txt_filename,
    )


def command_line_plot_magnon_dos():
    """
    map plot_magnon_dos to a shell command using argparse
    """
    parser = argparse.ArgumentParser(description="Plot magnon DOS")
    # parser.add_argument("fname", type=str, help="exchange.xml")
    parser.add_argument(
        "-p",
        "--path",
        type=str,
        default="./",
        help="path to exchange.xml, often the TB2J_results directory, or the Multibinit directory inside it.",
    )
    parser.add_argument(
        "-n", "--npoints", type=int, default=301, help="number of points in the energy"
    )
    parser.add_argument(
        "-w",
        "--window",
        type=float,
        nargs=2,
        default=None,
        help="energy window for the dos, two numbers giving the lower and upper bound",
    )
    parser.add_argument(
        "-k", "--kmesh", type=int, nargs=3, default=[9, 9, 9], help="k mesh"
    )

    parser.add_argument(
        "-s",
        "--smearing_width",
        type=float,
        default=10,
        help="Gauss smearing width in meV.",
    )

    parser.add_argument(
        "-g", "--gamma", action="store_true", help="Use gamma centered k mesh."
    )
    parser.add_argument(
        "-Jq",
        action="store_true",
        help="Plot the eigenvalues of J(q) instead of the magnon frequency.",
    )
    parser.add_argument(
        "--show",
        action="store_true",
        help="show the figure. By default it is not shown on the screen.",
    )
    parser.add_argument(
        "-f",
        "--fig_filename",
        type=str,
        default="magnon_dos.pdf",
        help="output filename for figure. Default: magnon_dos.pdf",
    )
    parser.add_argument(
        "-t",
        "--txt_filename",
        type=str,
        default="magnon_dos.txt",
        help="output filename of the data for the magnon DOS. Default: magnond_dos.txt",
    )
    args = parser.parse_args()
    plot_magnon_dos(
        # fname=args.fname,
        path=args.path,
        npoints=args.npoints,
        kmesh=args.kmesh,
        gamma=args.gamma,
        smearing_width=args.smearing_width / 1000,
        Jq=args.Jq,
        figfname=args.fig_filename,
        txt_filename=args.txt_filename,
        show=args.show,
    )



import gc
from pathlib import Path

import numpy as np
import tqdm
from HamiltonIO.abacus.abacus_wrapper import AbacusSplitSOCParser
from HamiltonIO.model.occupations import GaussOccupations
from typing_extensions import DefaultDict

from TB2J.anisotropy import Anisotropy
from TB2J.exchange import ExchangeNCL
from TB2J.external import p_map
from TB2J.mathutils.fibonacci_sphere import fibonacci_semisphere

# from HamiltonIO.model.rotate_spin import rotate_Matrix_from_z_to_axis, rotate_Matrix_from_z_to_sperical
# from TB2J.abacus.abacus_wrapper import AbacusWrapper, AbacusParser
from TB2J.mathutils.rotate_spin import (
    rotate_spinor_matrix,
)


def get_occupation(evals, kweights, nel, width=0.1):
    occ = GaussOccupations(nel=nel, width=width, wk=kweights, nspin=2)
    return occ.occupy(evals)


class MAEGreen(ExchangeNCL):
    def __init__(self, angles=None, **kwargs):
        """
        angles are defined as theta, phi pairs, where theta is the angle between the z-axis and the magnetization direction, and phi is the angle between the x-axis and the projection of the magnetization direction on the x-y plane.
        """
        super().__init__(**kwargs)
        self.natoms = len(self.atoms)
        if angles is None or angles == "xyz":
            self.set_angles_xyz()
        elif angles == "axis":
            self.set_angles_axis()
        elif angles == "scan":
            self.set_angles_scan()
        elif angles == "fib":
            self.set_angles_fib()
        elif angles == "random":
            self.set_angles_random()
        elif angles == "miller":
            self.set_angles_miller()
        else:
            self.thetas = angles[0]
            self.phis = angles[1]

        nangles = len(self.thetas)
        self.es = np.zeros(nangles, dtype=complex)
        self.es_atom = np.zeros((nangles, self.natoms), dtype=complex)
        self.es_atom_orb = DefaultDict(lambda: 0)

    def set_angles_xyz(self):
        """theta and phi are defined as the x, y, z, xy, yz, xz, xyz, x-yz, -xyz, -x-yz axis."""
        self.thetas = [np.pi / 2, np.pi / 2, 0.0]
        self.phis = [np.pi / 2, 0, 0.0]

    def set_angles_axis(self):
        """theta and phi are defined as the x, y, z, xy, yz, xz, xyz, x-yz, -xyz, -x-yz axis."""
        self.thetas = [0, np.pi / 2, np.pi / 2, np.pi / 2, np.pi, 0, np.pi / 2, 0, 0, 0]
        self.phis = [0, 0, np.pi / 2, np.pi / 4, 0, 0, 0, np.pi]

    def set_angles_miller(self, nmax=2):
        """theta and angles corresponding to the miller index. remove duplicates.
        e.g. 002 and 001 are the same.
        """
        thetas = []
        phis = []
        for k in range(0, nmax + 1):
            for j in range(-nmax, nmax + 1):
                for i in range(-nmax, nmax + 1):
                    if i == 0 and j == 0 and k == 0:
                        continue
                    thetas.append(np.arccos(k / np.sqrt(i**2 + j**2 + k**2)))
                    if i == 0 and j == 0:
                        phis.append(0)
                    else:
                        phis.append(np.arctan2(j, i))
        self.thetas = thetas
        self.phis = phis
        self.angle_pairs = list(zip(thetas, phis))
        self.angle_pairs = list(set(self.angle_pairs))
        self.thetas, self.phis = zip(*self.angle_pairs)

    def set_angles_scan(self, step=15):
        self.thetas = []
        self.phis = []
        for i in range(0, 181, step):
            for j in range(0, 181, step):
                self.thetas.append(i * np.pi / 180)
                self.phis.append(j * np.pi / 180)

    def set_angles_random(self, n=16):
        # n random pairs of theta, phi
        self.thetas = np.random.random(n) * np.pi
        self.phis = np.random.random(n) * 2 * np.pi

    def set_angles_fib(self, n=35):
        self.thetas, self.phis = fibonacci_semisphere(n)
        # thetas and phis are np.array
        # add (theta, phi): (pi/2, pi/2) and (pi/2, pi/4)
        # self.thetas += [np.pi / 2, np.pi / 2]
        # self.phis += [np.pi / 2, np.pi / 4]
        for i in range(8):
            self.thetas = np.concatenate([self.thetas, [np.pi / 2]])
            self.phis = np.concatenate([self.phis, [np.pi * i / 8]])

    def get_band_energy_vs_angles_from_eigen(
        self,
        thetas,
        phis,
    ):
        """
        Calculate the band energy for a given set of angles by using the eigenvalues and eigenvectors of the Hamiltonian.
        """
        es = []
        nangles = len(thetas)
        self.tbmodel.set_so_strength(1.0)
        for i in tqdm.trange(nangles):
            theta = thetas[i]
            phi = phis[i]
            self.tbmodel.set_Hsoc_rotation_angle([theta, phi])
            e = self.get_band_energy()
            es.append(e)
        return es

    def get_band_energy(self):
        self.width = 0.1
        evals, _evecs = self.tbmodel.solve_all(self.G.kpts)
        occ = get_occupation(evals, self.G.kweights, self.tbmodel.nel, width=self.width)
        eband = np.sum(evals * occ * self.G.kweights[:, np.newaxis])
        return eband

    def get_perturbed(self, e, thetas, phis):
        self.tbmodel.set_so_strength(0.0)
        # maxsoc = self.tbmodel.get_max_Hsoc_abs()
        # maxH0 = self.tbmodel.get_max_H0_spin_abs()
        # if maxsoc > maxH0 * 0.01:
        #    print(f"""Warning: The SOC of the Hamiltonian has a maximum of {maxsoc} eV,
        #          comparing to the maximum of {maxH0} eV of the spin part of the Hamiltonian.
        #          The SOC is too strong, the perturbation theory may not be valid.""")

        G0K = self.G.get_Gk_all(e)
        Hsoc_k = self.tbmodel.get_Hk_soc(self.G.kpts)
        na = len(thetas)
        dE_angle = np.zeros(na, dtype=complex)
        dE_angle_atom = np.zeros((na, self.natoms), dtype=complex)
        # dE_angle_orbitals = np.zeros((na, self.natoms, self.norb, self.norb), dtype=complex)
        # dE_angle_orbitals = DefaultDict(lambda: 0)
        dE_angle_atom_orb = DefaultDict(lambda: 0)
        for iangle, (theta, phi) in enumerate(zip(thetas, phis)):
            for ik, dHk in enumerate(Hsoc_k):
                dHi = rotate_spinor_matrix(dHk, theta, phi)
                GdH = G0K[ik] @ dHi
                # dE += np.trace(GdH @ G0K[i].T.conj() @ dHi) * self.kweights[i]
                # diagonal of second order perturbation.
                # dG2diag = np.diag(GdH @ GdH)
                # dG2 = np.einsum("ij,ji->ij", GdH,   GdH)
                dG2 = GdH * GdH.T
                dG2sum = np.sum(dG2)
                # print(f"dG2sum-sum: {dG2sum}")
                # dG2sum = np.sum(dG2diag)

                # dG2sum = np.trace(GdH @ GdH)
                # print(f"dG2sum-Tr: {dG2sum}")
                # dG1sum = np.trace(GdH)
                # print(f"dG1sum-Tr: {dG1sum}")

                # dG2diag = np.diag(GdH @G0K[i].T.conj() @ dHi)
                # dE_angle[iangle] += np.trace(GdH@GdH) * self.G.kweights[ik]
                # dE_angle[iangle] += np.trace(GdH@G0K[ik].T.conj()@dHi ) * self.G.kweights[ik]
                dE_angle[iangle] += dG2sum * self.G.kweights[ik]
                for iatom in range(self.natoms):
                    iorb = self.iorb(iatom)
                    # dG2= dG2[::2, ::2] + dG2[1::2, 1::2] + dG2[1::2, ::2] + dG2[::2, 1::2]
                    dE_atom_orb = dG2[np.ix_(iorb, iorb)] * self.G.kweights[ik]
                    dE_atom_orb = (
                        dE_atom_orb[::2, ::2]
                        + dE_atom_orb[1::2, 1::2]
                        + dE_atom_orb[1::2, ::2]
                        + dE_atom_orb[::2, 1::2]
                    )
                    dE_atom = np.sum(dE_atom_orb)
                    mmat = self.mmats[iatom]
                    dE_atom_orb = mmat.T @ dE_atom_orb @ mmat
                    dE_angle_atom_orb[(iangle, iatom)] += dE_atom_orb
                    # dE_atom = np.sum(dG2diag[iorb]) * self.G.kweights[ik]
                    dE_angle_atom[iangle, iatom] += dE_atom
        return dE_angle, dE_angle_atom, dE_angle_atom_orb

    def get_perturbed_R(self, e, thetas, phis):
        self.tbmodel.set_so_strength(0.0)
        # Here only the first R vector is considered.
        # TODO: consider all R vectors.
        # Rlist = np.zeros((1, 3), dtype=float)
        # G0K = self.G.get_Gk_all(e)
        # G0R = k_to_R(self.G.kpts, Rlist, G0K, self.G.kweights)
        # dE_angle = np.zeros(len(thetas), dtype=complex)
        # dE_angle_atoms = np.zeros((len(thetas), self.natoms), dtype=complex)
        pass

    def get_band_energy_vs_angles(self, thetas, phis, with_eigen=False):
        if with_eigen:
            self.es2 = self.get_band_energy_vs_angles_from_eigen(thetas, phis)

        # for ie, e in enumerate(tqdm.tqdm(self.contour.path)):
        #    dE_angle, dE_angle_atom, dE_angle_atom_orb = self.get_perturbed(
        #        e, thetas, phis
        #    )
        #    self.es += dE_angle * self.contour.weights[ie]
        #    self.es_atom += dE_angle_atom * self.contour.weights[ie]
        #    for key, value in dE_angle_atom_orb.items():
        #        self.es_atom_orb[key] += (
        #            dE_angle_atom_orb[key] * self.contour.weights[ie]
        #        )

        # rewrite the for loop above to use p_map
        def func(e):
            return self.get_perturbed(e, thetas, phis)

        if self.nproc > 1:
            results = p_map(func, self.contour.path, num_cpus=self.nproc)
        else:
            npole = len(self.contour.path)
            results = map(func, tqdm.tqdm(self.contour.path, total=npole))
        for i, result in enumerate(results):
            dE_angle, dE_angle_atom, dE_angle_atom_orb = result
            self.es += dE_angle * self.contour.weights[i]
            self.es_atom += dE_angle_atom * self.contour.weights[i]
            for key, value in dE_angle_atom_orb.items():
                self.es_atom_orb[key] += (
                    dE_angle_atom_orb[key] * self.contour.weights[i]
                )

        self.es = -np.imag(self.es) / (2 * np.pi)
        self.es_atom = -np.imag(self.es_atom) / (2 * np.pi)
        for key, value in self.es_atom_orb.items():
            self.es_atom_orb[key] = -np.imag(value) / (2 * np.pi)

    def fit_anisotropy_tensor(self):
        self.ani = Anisotropy.fit_from_data(self.thetas, self.phis, self.es)
        return self.ani

    def output(
        self,
        output_path="TB2J_anisotropy",
        with_eigen=True,
        figure3d="MAE_3d.png",
        figure_contourf="MAE_contourf.png",
    ):
        Path(output_path).mkdir(exist_ok=True)
        fname = f"{output_path}/MAE.dat"
        fname_orb = f"{output_path}/MAE_orb.dat"
        # fname_tensor = f"{output_path}/MAE_tensor.dat"
        # if figure3d is not None:
        #    fname_fig3d = f"{output_path}/{figure3d}"
        # if figure_contourf is not None:
        #    fname_figcontourf = f"{output_path}/{figure_contourf}"

        # ouput with eigenvalues.
        if with_eigen:
            fname_eigen = f"{output_path}/MAE_eigen.dat"
            with open(fname_eigen, "w") as f:
                f.write("# theta, phi, MAE(total), MAE(atom-wise) Unit: meV\n")
                for i, (theta, phi, e, es) in enumerate(
                    zip(self.thetas, self.phis, self.es2, self.es_atom)
                ):
                    f.write(f"{theta:.5f} {phi:.5f} {e*1e3:.8f} ")
                    for ea in es:
                        f.write(f"{ea*1e3:.8f} ")
                    f.write("\n")

        with open(fname, "w") as f:
            f.write("# theta (rad), phi(rad), MAE(total), MAE(atom-wise) Unit: meV\n")
            for i, (theta, phi, e, es) in enumerate(
                zip(self.thetas, self.phis, self.es, self.es_atom)
            ):
                f.write(f"{theta%np.pi:.5f} {phi%(2*np.pi):.5f} {e*1e3:.8f} ")
                for ea in es:
                    f.write(f"{ea*1e3:.8f} ")
                f.write("\n")

        # self.ani = self.fit_anisotropy_tensor()
        # with open(fname_tensor, "w") as f:
        #    f.write("# Anisotropy tensor in meV\n")
        #    f.write(f"{self.ani.tensor_strings(include_isotropic=False)}\n")

        # if figure3d is not None:
        #    self.ani.plot_3d(figname=fname_fig3d, show=False)

        # if figure_contourf is not None:
        #    self.ani.plot_contourf(figname=fname_figcontourf, show=False)

        # plt.close()
        gc.collect()

        with open(fname_orb, "w") as f:
            f.write("=" * 80 + "\n")
            f.write("Orbitals for each atom: \n")
            f.write("=" * 80 + "\n")
            f.write("Note: the energies are in meV\n")
            for iatom in range(self.natoms):
                f.write(f"Atom {iatom:03d}: ")
                for orb in self.orbital_names[iatom]:
                    f.write(f"{orb} ")
                f.write("\n")
            for i, (theta, phi, e, eatom) in enumerate(
                zip(self.thetas, self.phis, self.es, self.es_atom)
            ):
                f.write("-" * 60 + "\n")
                f.write(f"Angle {i:03d}:   theta={theta:.5f} phi={phi:.5f} \n ")
                f.write(f"E: {e*1e3:.8f} \n")
                for iatom, ea in enumerate(eatom):
                    f.write(f"Atom {iatom:03d}: {ea*1e3:.8f} \n")
                    f.write("Orbital: ")
                    eorb = self.es_atom_orb[(i, iatom)]

                    # write numpy matrix to file
                    f.write(
                        np.array2string(
                            eorb * 1e3, precision=4, separator=",", suppress_small=True
                        )
                    )

                    eorb_diff = eorb - self.es_atom_orb[(0, iatom)]
                    f.write("Diference to the first angle: ")
                    f.write(
                        np.array2string(
                            eorb_diff * 1e3,
                            precision=4,
                            separator=",",
                            suppress_small=True,
                        )
                    )
                f.write("\n")

    def run(self, output_path="TB2J_anisotropy", with_eigen=False):
        self.get_band_energy_vs_angles(self.thetas, self.phis, with_eigen=with_eigen)
        self.output(output_path=output_path, with_eigen=with_eigen)


def abacus_get_MAE(
    path_nosoc,
    path_soc,
    kmesh,
    thetas,
    phis,
    gamma=True,
    output_path="TB2J_anisotropy",
    magnetic_elements=None,
    nel=None,
    width=0.1,
    with_eigen=False,
    **kwargs,
):
    """Get MAE from Abacus with magnetic force theorem. Two calculations are needed. First we do an calculation with SOC but the soc_lambda is set to 0. Save the density. The next calculatin we start with the density from the first calculation and set the SOC prefactor to 1. With the information from the two calcualtions, we can get the band energy with magnetic moments in the direction, specified in two list, thetas, and phis."""
    parser = AbacusSplitSOCParser(
        outpath_nosoc=path_nosoc, outpath_soc=path_soc, binary=False
    )
    model = parser.parse()
    model.set_so_strength(0.0)
    if nel is not None:
        model.nel = nel
    mae = MAEGreen(
        tbmodels=model,
        atoms=model.atoms,
        kmesh=kmesh,
        efermi=None,
        basis=model.basis,
        angles=[thetas, phis],
        magnetic_elements=magnetic_elements,
        **kwargs,
    )
    mae.run(output_path=output_path, with_eigen=with_eigen)



import os

import numpy as np

from TB2J.exchange import ExchangeNCL
from TB2J.exchangeCL2 import ExchangeCL2
from TB2J.io_merge import merge
from TB2J.MAEGreen import MAEGreen

try:
    from HamiltonIO.siesta import SislParser
except ImportError:
    print(
        "Cannot import SislWrapper from HamiltonIO.siesta. Please install HamiltonIO first."
    )


def siesta_anisotropy(**kwargs):
    pass


def gen_exchange_siesta(fdf_fname, read_H_soc=False, **kwargs):
    """
    parameters:
        fdf_fname: str
            The fdf file for the calculation.
        read_H_soc: bool
            Whether to read the SOC Hamiltonian. Default is False.

    parameters in **kwargs:
        magnetic_elements: list of str
            The magnetic elements to calculate the exchange. e.g. ["Fe", "Co", "Ni"]
        include_orbs: dict
            The included orbitals for each magnetic element. e.g. {"Fe": ["d"]}. Default is None.
        kmesh: list of int
            The k-point mesh for the calculation. e.g. [5, 5, 5]. Default is [5, 5, 5].
        emin: float
            The minimum energy for the calculation. Default is -12.0.
        emax: float
            The maximum energy for the calculation. Default is 0.0.
        nz: int
            The number of points for the energy mesh. Default is 100.
        exclude_orbs: list of str
            The excluded orbitals for the calculation. Default is [].
        Rcut: float
            The cutoff radius for the exchange calculation in angstrom. Default is None.
        ne: int
            The number of electrons for the calculation. Default is None.
        nproc: int
            The number of processors for the calculation. Default is 1.
        use_cache: bool
            Whether to use the cache for the calculation. Default is False.
        output_path: str
            The output path for the calculation. Default is "TB2J_results".
        orb_decomposition: bool
            Whether to calculate the orbital decomposition. Default is False.
        orth: bool
            Whether to orthogonalize the orbitals. Default is False.
        description: str
            The description for the calculation. Default is "".
    """

    exargs = dict(
        magnetic_elements=[],
        include_orbs=None,
        kmesh=[5, 5, 5],
        emin=-12.0,
        emax=0.0,
        nz=100,
        exclude_orbs=[],
        Rcut=None,
        ne=None,
        nproc=1,
        use_cache=False,
        output_path="TB2J_results",
        orb_decomposition=False,
        orth=False,
        ibz=False,
        description="",
    )
    exargs.update(kwargs)
    try:
        import sisl
    except ImportError:
        raise ImportError("sisl cannot be imported. Please install sisl first.")

    from packaging import version

    if version.parse(sisl.__version__) <= version.parse("0.10.0"):
        raise ImportError(
            f"sisl version is {sisl.__version__}, but should be larger than 0.10.0."
        )

    output_path = exargs.pop("output_path")

    parser = SislParser(
        fdf_fname=fdf_fname, ispin=None, read_H_soc=read_H_soc, orth=exargs["orth"]
    )
    if parser.spin.is_colinear:
        print("Reading Siesta hamiltonian: colinear spin.")
        # tbmodel_up = SislWrapper(fdf_fname=None, sisl_hamiltonian=H, spin=0, geom=geom)
        # tbmodel_dn = SislWrapper(fdf_fname=None, sisl_hamiltonian=H, spin=1, geom=geom)
        tbmodel_up, tbmodel_dn = parser.get_model()
        basis = dict(zip(tbmodel_up.orbs, list(range(tbmodel_up.norb))))
        print("Starting to calculate exchange.")
        description = f""" Input from collinear Siesta data.
 working directory: {os.getcwd()}
 fdf_fname: {fdf_fname}.
\n"""
        exchange = ExchangeCL2(
            tbmodels=(tbmodel_up, tbmodel_dn),
            atoms=tbmodel_up.atoms,
            basis=basis,
            efermi=0.0,
            # magnetic_elements=exargs['magnetic_elements'],
            # include_orbs=ex
            **exargs,
        )
        exchange.run(path=output_path)
        print("\n")
        print(f"All calculation finished. The results are in {output_path} directory.")

    elif parser.spin.is_spinorbit or parser.spin.is_noncolinear:
        print("Reading Siesta hamiltonian: non-colinear spin.")
        model = parser.get_model()
        basis = dict(zip(model.orbs, list(range(model.nbasis))))
        print("Starting to calculate exchange.")
        description = f""" Input from non-collinear Siesta data.
 working directory: {os.getcwd()}
 fdf_fname: {fdf_fname}.
Warning: The DMI component parallel to the spin orientation, the Jani which has the component of that orientation should be disregarded
 e.g. if the spins are along z, the xz, yz, zz, zx, zy components and the z component of DMI.
 If you need these component, try to do three calculations with spin along x, y, z,  or use structure with z rotated to x, y and z. And then use TB2J_merge.py to get the full set of parameters.
\n"""
        exargs["description"] = description
        if not model.split_soc:
            exchange = ExchangeNCL(
                tbmodels=model,
                atoms=model.atoms,
                basis=basis,
                efermi=0.0,
                # magnetic_elements=magnetic_elements,
                # include_orbs=include_orbs,
                output_path=output_path,
                **exargs,
            )
            exchange.run(path=output_path)
            print("\n")
            print(
                f"All calculation finished. The results are in {output_path} directory."
            )
        else:
            print("Starting to calculate MAE.")
            model.set_so_strength(0.0)
            MAE = MAEGreen(
                tbmodels=model,
                atoms=model.atoms,
                basis=basis,
                efermi=None,
                angles="axis",
                # magnetic_elements=magnetic_elements,
                # include_orbs=include_orbs,
                **exargs,
            )
            # thetas = [0, np.pi / 2, np.pi, 3 * np.pi / 2]
            # phis = [0, 0, 0, 0]
            # MAE.set_angles(thetas=thetas, phis=phis)
            MAE.run(output_path=f"{output_path}_anisotropy", with_eigen=False)
            # print(
            #    f"MAE calculation finished. The results are in {output_path} directory."
            # )

            angle = {"x": (np.pi / 2, 0), "y": (np.pi / 2, np.pi / 2), "z": (0, 0)}
            for key, val in angle.items():
                theta, phi = val
                model.set_so_strength(1.0)
                model.set_Hsoc_rotation_angle([theta, phi])
                basis = dict(zip(model.orbs, list(range(model.nbasis))))
                output_path_full = f"{output_path}_{key}"
                exchange = ExchangeNCL(
                    tbmodels=model,
                    atoms=model.atoms,
                    basis=basis,
                    efermi=None,  # set to None, compute from efermi.
                    **exargs,
                )
                exchange.run(path=output_path_full)
                print("\n")
                print(
                    f"All calculation finished. The results are in {output_path_full} directory."
                )

            merge(
                "TB2J_results_x",
                "TB2J_results_y",
                "TB2J_results_z",
                main_path=None,
                save=True,
                write_path="TB2J_results",
            )
            print("Final TB2J_results written to TB2J_results directory.")



import numpy as np
from TB2J.pauli import pauli_block_all, pauli_block_sigma_norm
from TB2J.io_exchange import SpinIO
from functools import lru_cache
from TB2J.exchange import ExchangeNCL
from collections import defaultdict


class ExchangePert(ExchangeNCL):
    def set_dHdx(self, dHdx):
        self.dHdx = dHdx
        self.dHdxR0 = dHdx.ham_R0
        assert self.dHdx.nbasis == self.nbasis
        self.AO_ijR = defaultdict(lambda: np.zeros((4, 4), dtype=complex))
        self.AT_ijR = defaultdict(lambda: np.zeros((4, 4), dtype=complex))

    @lru_cache()
    def get_dP_iatom(self, iatom):
        orbs = self.iorb(iatom)
        return pauli_block_sigma_norm(self.dHdxR0[np.ix_(orbs, orbs)])

    def get_A_ijR(self, G, dG, iatom, jatom, de):
        """calculate A from G for a energy slice (de).
        It take the
        .. math::
           A^{uv} = p T^u p T^v dE / pi

        where u, v are I, x, y, z (index 0, 1,2,3). p(i) = self.get_P_iatom(iatom)
        T^u(ijR)  (u=0,1,2,3) = pauli_block_all(G)

        :param G: Green's function for all R, i, j.
        :param iatom: i
        :param jatom: j
        :param de:  energy step. used for integeration
        :returns: a matrix of A_ij(u, v), where u, v =(0)0, x(1), y(2), z(3)
        :rtype:  4*4 matrix
        """

        for R in self.Rlist:
            # G[i, j, R]
            GR = G[R]
            Gij = self.GR_atom(GR, iatom, jatom)
            # GijR , I, x, y, z component.
            Gij_Ixyz = pauli_block_all(Gij)

            # G(j, i, -R)
            Rm = tuple(-x for x in R)
            GRm = G[Rm]
            Gji = self.GR_atom(GRm, jatom, iatom)
            Gji_Ixyz = pauli_block_all(Gji)

            dGR = dG[R]
            dGij = self.GR_atom(dGR, iatom, jatom)
            # GijR , I, x, y, z component.
            dGij_Ixyz = pauli_block_all(dGij)

            # G(j, i, -R)
            Rm = tuple(-np.array(R))
            dGRm = dG[Rm]
            dGji = self.GR_atom(dGRm, jatom, iatom)
            dGji_Ixyz = pauli_block_all(dGji)

            tmp1 = np.zeros((4, 4), dtype=complex)
            tmp2 = np.zeros((4, 4), dtype=complex)
            tmp3 = np.zeros((4, 4), dtype=complex)
            for a in range(4):
                for b in range(4):
                    AijRab = np.matmul(
                        np.matmul(self.get_P_iatom(iatom), Gij_Ixyz[a]),
                        np.matmul(self.get_P_iatom(jatom), Gji_Ixyz[b]),
                    )

                    A1 = np.matmul(
                        np.matmul(self.get_P_iatom(iatom), dGij_Ixyz[a]),
                        np.matmul(self.get_P_iatom(jatom), Gji_Ixyz[b]),
                    )
                    A2 = np.matmul(
                        np.matmul(self.get_P_iatom(iatom), Gij_Ixyz[a]),
                        np.matmul(self.get_P_iatom(jatom), dGji_Ixyz[b]),
                    )
                    A3 = np.matmul(
                        np.matmul(self.get_dP_iatom(iatom), Gij_Ixyz[a]),
                        np.matmul(self.get_P_iatom(jatom), Gji_Ixyz[b]),
                    )
                    A4 = np.matmul(
                        np.matmul(self.get_P_iatom(iatom), Gij_Ixyz[a]),
                        np.matmul(self.get_dP_iatom(jatom), Gji_Ixyz[b]),
                    )
                    AOijRab = A1 + A2 + A3 + A4

                    if False:
                        B1 = np.matmul(
                            np.matmul(self.get_dP_iatom(iatom), dGij_Ixyz[a]),
                            np.matmul(self.get_P_iatom(jatom), Gji_Ixyz[b]),
                        )
                        B2 = np.matmul(
                            np.matmul(self.get_dP_iatom(iatom), Gij_Ixyz[a]),
                            np.matmul(self.get_dP_iatom(jatom), Gji_Ixyz[b]),
                        )
                        B3 = np.matmul(
                            np.matmul(self.get_dP_iatom(iatom), Gij_Ixyz[a]),
                            np.matmul(self.get_P_iatom(jatom), dGji_Ixyz[b]),
                        )
                        B4 = np.matmul(
                            np.matmul(self.get_P_iatom(iatom), dGij_Ixyz[a]),
                            np.matmul(self.get_dP_iatom(jatom), Gji_Ixyz[b]),
                        )
                        B5 = np.matmul(
                            np.matmul(self.get_P_iatom(iatom), dGij_Ixyz[a]),
                            np.matmul(self.get_P_iatom(jatom), dGji_Ixyz[b]),
                        )

                        B6 = np.matmul(
                            np.matmul(self.get_P_iatom(iatom), Gij_Ixyz[a]),
                            np.matmul(self.get_dP_iatom(jatom), dGji_Ixyz[b]),
                        )
                        tmp3[a, b] = np.trace(B1 + B2 + B3 + B4 + B5 + B6)

                    # trace over orb
                    tmp1[a, b] = np.trace(AijRab)
                    tmp2[a, b] = np.trace(AOijRab)
            # Note: the full complex, rather than Re or Im part is stored into A_ijR.
            self.A_ijR[(R, iatom, jatom)] += tmp1 * de / np.pi
            self.AO_ijR[(R, iatom, jatom)] += tmp2 * de / np.pi
            self.AT_ijR[(R, iatom, jatom)] += tmp3 * de / np.pi

    def get_all_A(self, G, dG, de):
        """
        Calculate all A matrix elements
        Loop over all magnetic atoms.
        :param G: Green's function.
        :param de: energy step.
        """
        for iatom in self.ind_mag_atoms:
            for jatom in self.ind_mag_atoms:
                self.get_A_ijR(G, dG, iatom, jatom, de)

    def A_to_Jtensor(self):
        """
        Calculate J tensors from A.
        If we assume the exchange can be written as a bilinear tensor form,
        J_{isotropic} = Tr Im (A^{00} - A^{xx} - A^{yy} - A^{zz})
        """
        super().A_to_Jtensor()
        self.dJdx = {}
        for key, val in self.AO_ijR.items():
            # key:(R, iatom, jatom)
            R, iatom, jatom = key
            ispin = self.ispin(iatom)
            jspin = self.ispin(jatom)
            keyspin = (R, ispin, jspin)
            is_nonself = not (R == (0, 0, 0) and iatom == jatom)
            dJiso = np.zeros((3, 3), dtype=float)
            # Heisenberg like J.
            for i in range(3):
                dJiso[i, i] += np.imag(val[0, 0] - val[3, 3])
            if is_nonself:
                self.dJdx[keyspin] = dJiso[0, 0]

        self.dJdx2 = {}
        for key, val in self.AT_ijR.items():
            # key:(R, iatom, jatom)
            R, iatom, jatom = key
            ispin = self.ispin(iatom)
            jspin = self.ispin(jatom)
            keyspin = (R, ispin, jspin)
            is_nonself = not (R == (0, 0, 0) and iatom == jatom)
            dJ2iso = np.zeros((3, 3), dtype=float)
            # Heisenberg like J.
            for i in range(3):
                dJ2iso[i, i] += np.imag(val[0, 0] - val[3, 3])
            if is_nonself:
                self.dJdx2[keyspin] = dJ2iso[0, 0]

    def calculate_all(self):
        """
        The top level.
        """
        print("Green's function Calculation started.")

        # widgets = [
        #    " [",
        #    progressbar.Timer(),
        #    "] ",
        #    progressbar.Bar(),
        #    " (",
        #    progressbar.ETA(),
        #    ") ",
        # ]
        # bar = progressbar.ProgressBar(maxval=self.contour.npoints, widgets=widgets)
        # bar.start()
        for ie in range(self.contour.npoints):
            # bar.update(ie)
            e = self.contour.elist[ie]
            de = self.contour.de[ie]
            GR, dGdx = self.G.get_GR_and_dGRdx(self.Rlist, energy=e, dHdx=self.dHdx)
            # self.get_rho_e(GR, de)
            self.get_all_A(GR, dGdx, de)
            if self.calc_NJt:
                self.get_N_e(GR, de)

        self.get_rho_atom()
        self.A_to_Jtensor()
        if self.calc_NJt:
            self.calculate_DMI_NJT()
        # bar.finish()

    def write_output(self, path="TB2J_results"):
        self._prepare_index_spin()
        output = SpinIO(
            atoms=self.atoms,
            charges=self.charges,
            spinat=self.spinat,
            index_spin=self.index_spin,
            colinear=True,
            distance_dict=self.distance_dict,
            exchange_Jdict=self.exchange_Jdict,
            dJdx=self.dJdx,
            dmi_ddict=None,
            NJT_Jdict=None,
            NJT_ddict=None,
            bilinear_Jdict=None,
            biquadratic_Jdict=self.B,
        )
        output.write_all(path=path)



import numpy as np

__all__ = ["generate_grid", "get_rotation_arrays", "round_to_precision", "uz", "I"]

I = np.eye(3)
uz = np.array([[0.0, 0.0, 1.0]])


def generate_grid(kmesh, sort=True):
    half_grid = [int(n / 2) for n in kmesh]
    grid = np.stack(
        np.meshgrid(*[np.arange(-n, n + 1) for n in half_grid]), axis=-1
    ).reshape(-1, 3)

    if sort:
        idx = np.linalg.norm(grid, axis=-1).argsort()
        grid = grid[idx]

    return grid


# def JR_to_Jq(JR, Rlist, qpt, vecn):
#    for iR, R in enumerate(Rlist):
#        phase = 2 * np.pi * R @ qpt
#        rv = phase * vecn
#        Rot = Rotation.from_rotvec(rv.reshape(-1, 3)).as_matrix().reshape(R.shape[0], 3, 3)
#


def get_rotation_arrays(magmoms, u=uz):
    dim = magmoms.shape[0]
    v = magmoms
    n = np.cross(u, v)
    n /= np.linalg.norm(n, axis=-1).reshape(dim, 1)
    z = np.repeat(u, dim, axis=0)
    A = np.stack([z, np.cross(n, z), n], axis=1)
    B = np.stack([v, np.cross(n, v), n], axis=1)
    R = np.einsum("nki,nkj->nij", A, B)

    Rnan = np.isnan(R)
    if Rnan.any():
        nanidx = np.where(Rnan)[0]
        R[nanidx] = I
        R[nanidx, 2] = v[nanidx]

    U = R[:, 0] + 1j * R[:, 1]
    V = R[:, 2]

    return U, V


def round_to_precision(array, precision):
    return precision * np.round(array / precision)



import os
import numpy as np
import copy
from scipy.linalg import eigh
from scipy.sparse import csr_matrix
from scipy.io import netcdf_file
from collections import defaultdict

# from tbmodels import Model
from ase.atoms import Atoms
from TB2J.utils import auto_assign_basis_name
from TB2J.wannier import parse_ham, parse_xyz, parse_atoms, parse_tb


class AbstractTB:
    def __init__(self, R2kfactor, nspin, norb):
        #: :math:`\alpha` used in :math:`H(k)=\sum_R  H(R) \exp( \alpha k \cdot R)`,
        #: Should be :math:`2\pi i` or :math:`-2\pi i`
        self.is_orthogonal = True
        self.R2kfactor = R2kfactor

        #: number of spin. 1 for collinear, 2 for spinor.
        self.nspin = nspin

        #:number of orbitals. Each orbital can have two spins.
        self.norb = norb

        #: nbasis=nspin*norb
        self.nbasis = nspin * norb

        #: The array of cartesian coordinate of all basis. shape:nbasis,3
        self.xcart = None

        #: The array of cartesian coordinate of all basis. shape:nbasis,3
        self.xred = None

        #: The order of the spinor basis.
        #: 1: orb1_up, orb2_up,  ... orb1_down, orb2_down,...
        #: 2: orb1_up, orb1_down, orb2_up, orb2_down,...

        self._name = None

    @property
    def name(self):
        return self._name

    def get_hamR(self, R):
        """
        get the Hamiltonian H(R), array of shape (nbasis, nbasis)
        """
        raise NotImplementedError()

    def get_orbs(self):
        """
        returns the orbitals.
        """
        raise NotImplementedError()

    def HSE(self, kpt):
        raise NotImplementedError()

    def HS_and_eigen(self, kpts):
        """
        get Hamiltonian, overlap matrices, eigenvalues, eigen vectors for all kpoints.

        :param:

        * kpts: list of k points.

        :returns:

        * H, S, eigenvalues, eigenvectors for all kpoints
        * H: complex array of shape (nkpts, nbasis, nbasis)
        * S: complex array of shape (nkpts, nbasis, nbasis). S=None if the basis set is orthonormal.
        * evals: complex array of shape (nkpts, nbands)
        * evecs: complex array of shape (nkpts, nbasis, nbands)
        """
        raise NotImplementedError()


class MyTB(AbstractTB):
    def __init__(
        self,
        nbasis,
        data=None,
        R_degens=None,
        positions=None,
        sparse=False,
        ndim=3,
        nspin=1,
        double_site_energy=2.0,
    ):
        """
        :param nbasis: number of basis.
        :param data: a dictionary of {R: matrix}. R is a tuple, matrix is a nbasis*nbasis matrix.
        :param R_degens: degeneracy of R.
        :param positions: reduced positions.
        :param sparse: Bool, whether to use a sparse matrix.
        :param ndim: number of dimensions.
        :param nspin: number of spins.
        """

        if data is not None:
            self.data = data
        else:
            self.data = defaultdict(lambda: np.zeros((nbasis, nbasis), dtype=complex))
        self._nbasis = nbasis
        self._nspin = nspin
        self._norb = nbasis // nspin
        self._ndim = ndim
        if R_degens is not None:
            self.R_degens = R_degens
        else:
            self.R_degens = np.ones(len(self.data.keys()), dtype=int)
        if positions is None:
            self._positions = np.zeros((nbasis, self.ndim))
        else:
            self._positions = positions
        self.prepare_phase_rjri()
        self.sparse = sparse
        self.double_site_energy = double_site_energy
        if sparse:
            self._matrix = csr_matrix
        self.atoms = None
        self.R2kfactor = 2.0j * np.pi
        self.k2Rfactor = -2.0j * np.pi
        self.is_orthogonal = True
        self._name = "Wannier"

    def set_atoms(self, atoms):
        self.atoms = atoms

    @property
    def nspin(self):
        return self._nspin

    @property
    def norb(self):
        """
        norb: number of orbitals, if spin/spinor, norb=nbasis/2
        """
        return self._norb

    @property
    def nbasis(self):
        return self._nbasis

    @property
    def ndim(self):
        return self._ndim

    @property
    def xcart(self):
        raise NotImplementedError()

    @property
    def xred(self):
        return self._positions

    @property
    def positions(self):
        return self._positions

    @property
    def onsite_energies(self):
        return self.data[(0, 0, 0)].diagonal() * 2

    @property
    def hoppings(self):
        """
        The hopping parameters, not including any onsite energy.
        """
        data = copy.deepcopy(self.data)
        np.fill_diagonal(data[(0, 0, 0)], 0.0)
        return data

    @staticmethod
    def read_from_wannier_dir(path, prefix, atoms=None, nls=True, groupby=None):
        """
        read tight binding model from a wannier function directory.
        :param path: path
        :param prefix: prefix to the wannier files, often wannier90, or wannier90_up, or wannier90_dn for vasp.
        """

        tb_fname = os.path.join(path, prefix + "_tb.dat")
        hr_fname = os.path.join(path, prefix + "_hr.dat")
        if os.path.exists(tb_fname):
            xcart, nbasis, data, R_degens = parse_tb(fname=tb_fname)
        else:
            nbasis, data, R_degens = parse_ham(
                fname=os.path.join(path, prefix + "_hr.dat")
            )
            xcart, _, _ = parse_xyz(fname=os.path.join(path, prefix + "_centres.xyz"))

        if atoms is None:
            atoms = parse_atoms(os.path.join(path, prefix + ".win"))
        cell = atoms.get_cell()
        xred = cell.scaled_positions(xcart)
        if groupby == "spin":
            norb = nbasis // 2
            xtmp = copy.deepcopy(xred)
            xred[::2] = xtmp[:norb]
            xred[1::2] = xtmp[norb:]
            for key, val in data.items():
                dtmp = copy.deepcopy(val)
                data[key][::2, ::2] = dtmp[:norb, :norb]
                data[key][::2, 1::2] = dtmp[:norb, norb:]
                data[key][1::2, ::2] = dtmp[norb:, :norb]
                data[key][1::2, 1::2] = dtmp[norb:, norb:]
        ind, positions = auto_assign_basis_name(xred, atoms)
        m = MyTB(nbasis=nbasis, data=data, positions=xred, R_degens=R_degens)
        nm = m.shift_position(positions)
        nm.set_atoms(atoms)
        return nm

    @staticmethod
    def load_banddownfolder(path, prefix, atoms=None, nls=True, groupby="spin"):
        from banddownfolder.scdm.lwf import LWF

        lwf = LWF.load_nc(fname=os.path.join(path, f"{prefix}.nc"))
        nbasis = lwf.nwann
        nspin = 1
        positions = lwf.wann_centers
        ndim = lwf.ndim
        H_mnR = defaultdict(lambda: np.zeros((nbasis, nbasis), dtype=complex))

        for iR, R in enumerate(lwf.Rlist):
            R = tuple(R)
            val = lwf.HwannR[iR]
            if np.linalg.norm(R) < 0.001:
                H_mnR[R] = val / 2.0
                # H_mnR[R] -= np.diag(np.diag(val) / 2.0)
            else:
                H_mnR[R] = val / 2.0
        m = MyTB(nbasis, data=H_mnR, nspin=nspin, ndim=ndim, positions=positions)
        m.atoms = atoms
        return m

    def gen_ham(self, k, convention=2):
        """
        generate hamiltonian matrix at k point.
        H_k( i, j)=\sum_R H_R(i, j)^phase.
        There are two conventions,
        first:
        phase =e^{ik(R+rj-ri)}. often better used for berry phase.
        second:
        phase= e^{ikR}. We use the first convention here.

        :param k: kpoint
        :param convention: 1 or 2.
        """
        Hk = np.zeros((self.nbasis, self.nbasis), dtype="complex")
        if convention == 2:
            for iR, (R, mat) in enumerate(self.data.items()):
                phase = np.exp(self.R2kfactor * np.dot(k, R))  # /self.R_degens[iR]
                H = mat * phase
                Hk += H + H.conjugate().T
        elif convention == 1:
            for iR, (R, mat) in enumerate(self.data.items()):
                phase = (
                    np.exp(self.R2kfactor * np.dot(k, R + self.rjminusri))
                    / self.R_degens[iR]
                )
                H = mat * phase
                Hk += H + H.conjugate().T
        else:
            raise ValueError("convention should be either 1 or 2.")
        return Hk

    def solve(self, k, convention=2):
        Hk = self.gen_ham(k, convention=convention)
        return eigh(Hk)

    def HSE_k(self, kpt, convention=2):
        H = self.gen_ham(tuple(kpt), convention=convention)
        S = None
        evals, evecs = eigh(H)
        return H, S, evals, evecs

    def HS_and_eigen(self, kpts, convention=2):
        """
        calculate eigens for all kpoints.
        :param kpts: list of k points.
        """
        nk = len(kpts)
        hams = np.zeros((nk, self.nbasis, self.nbasis), dtype=complex)
        evals = np.zeros((nk, self.nbasis), dtype=float)
        evecs = np.zeros((nk, self.nbasis, self.nbasis), dtype=complex)
        for ik, k in enumerate(kpts):
            hams[ik], S, evals[ik], evecs[ik] = self.HSE_k(
                tuple(k), convention=convention
            )
        return hams, None, evals, evecs

    def prepare_phase_rjri(self):
        """
        The matrix P: P(i, j) = r(j)-r(i)
        """
        self.rjminusri = self.xred[None, :, :] - self.xred[:, None, :]

    def to_sparse(self):
        for key, val in self.data:
            self.data[key] = self._matrix(val)

    @property
    def Rlist(self):
        return list(self.data.keys())

    @property
    def nR(self):
        """
        number of R
        """
        return len(self.Rlist)

    @property
    def site_energies(self):
        """
        on site energies.
        """
        return self.data[(0, 0, 0)].diagonal() * 2

    @property
    def ham_R0(self):
        """
        return hamiltonian at R=0. Note that the data is halfed for R=0.
        """
        return self.data[(0, 0, 0)] + self.data[(0, 0, 0)].T.conj()

    def get_hamR(self, R):
        """
        return the hamiltonian at H(i, j) at R.
        """
        nzR = np.nonzero(R)[0]
        if len(nzR) != 0 and R[nzR[0]] < 0:
            newR = tuple(-np.array(R))
            return self.data[newR].T.conj()
        elif len(nzR) == 0:
            newR = R
            mat = self.data[newR]
            return mat + self.data[(0, 0, 0)].T.conj()
        else:
            newR = R
            return self.data[newR]

    @staticmethod
    def _positive_R_mat(R, mat):
        nzR = np.nonzero(R)[0]
        if len(nzR) != 0 and R[nzR[0]] < 0:
            newR = tuple(np.array(-R))
            newmat = mat.T.conj()
        elif len(nzR) == 0:
            newR = R
            newmat = (mat + mat.T.conj()) / 2.0
        else:
            newR = R
            newmat = mat
        return newR, newmat

    def _to_positive_R(self):
        """
        make all the R positive.
        t(i, j, R) = t(j, i, -R).conj() if R is negative.
        """
        new_MyTB = MyTB(self.nbasis, sparse=self.sparse)
        for R, mat in self.data:
            newR, newmat = self._positive_R_mat(R, mat)
            new_MyTB[newR] += newmat
        return new_MyTB

    def shift_position(self, rpos):
        """
        shift the positions of basis set to near reference positions.
        E.g. reduced position 0.8, with refernce 0.0 will goto -0.2.
        This can move the wannier functions to near the ions.
        """
        pos = self.positions
        shift = np.zeros((self.nbasis, self.ndim), dtype="int")
        shift[:, :] = np.round(pos - rpos)
        newpos = copy.deepcopy(pos)
        for i in range(self.nbasis):
            newpos[i] = pos[i] - shift[i]
        d = MyTB(self.nbasis, ndim=self.ndim, nspin=self.nspin, R_degens=self.R_degens)
        d._positions = newpos

        for R, v in self.data.items():
            for i in range(self.nbasis):
                for j in range(self.nbasis):
                    sR = tuple(np.array(R) - shift[i] + shift[j])
                    nzR = np.nonzero(sR)[0]
                    if len(nzR) != 0 and sR[nzR[0]] < 0:
                        newR = tuple(-np.array(sR))
                        d.data[newR][j, i] += v[i, j].conj()
                    elif len(nzR) == 0:
                        newR = sR
                        d.data[newR][i, j] += v[i, j] * 0.5
                        d.data[newR][j, i] += v[i, j].conj() * 0.5
                    else:
                        d.data[sR][i, j] += v[i, j]
        return d

    def save(self, fname):
        """
        Save model into a netcdf file.
        :param fname: filename.
        """
        # from netCDF4 import Dataset
        # root = Dataset(fname, 'w', format="NETCDF4")
        root = netcdf_file(fname, mode="w")
        root.createDimension("nR", self.nR)
        root.createDimension("ndim", self.ndim)
        root.createDimension("nbasis", self.nbasis)
        root.createDimension("nspin", self.nspin)
        root.createDimension("natom", len(self.atoms))
        R = root.createVariable("R", "i4", ("nR", "ndim"))
        data_real = root.createVariable("data_real", "f8", ("nR", "nbasis", "nbasis"))
        data_imag = root.createVariable("data_imag", "f8", ("nR", "nbasis", "nbasis"))
        positions = root.createVariable("positions", "f8", ("nbasis", "ndim"))

        if self.atoms is not None:
            atom_numbers = root.createVariable("atom_numbers", "i4", ("natom",))
            atom_xred = root.createVariable("atom_xred", "f8", ("natom", "ndim"))
            atom_cell = root.createVariable("atom_cell", "f8", ("ndim", "ndim"))

        atom_cell.unit = "Angstrom"
        positions.unit = "1"
        data_real.unit = "eV"
        data_imag.unit = "eV"

        R[:] = np.array(self.Rlist)
        d = np.array(tuple(self.data.values()))
        data_real[:] = np.real(d)
        data_imag[:] = np.imag(d)
        positions[:] = np.array(self.positions)

        if self.atoms is not None:
            atom_numbers[:] = np.array(self.atoms.get_atomic_numbers())
            atom_xred[:] = np.array(self.atoms.get_scaled_positions(wrap=False))
            atom_cell[:] = np.array(self.atoms.get_cell())
        root.close()

    @staticmethod
    def load_MyTB(fname):
        """
        Load from a netcdf file.
        :param fname: netcdf filename.
        :Returns: tight binding model.
        """
        # from netCDF4 import Dataset
        root = netcdf_file(fname, "r", mmap=False)
        nbasis = root.dimensions["nbasis"]
        nspin = root.dimensions["nspin"]
        ndim = root.dimensions["ndim"]
        natom = root.dimensions["natom"]
        Rlist = root.variables["R"][:]
        mdata_real = root.variables["data_real"][:]
        mdata_imag = root.variables["data_imag"][:]
        positions = root.variables["positions"][:]

        atom_numbers = root.variables["atom_numbers"][:]
        atom_xred = root.variables["atom_xred"][:]
        atom_cell = root.variables["atom_cell"][:]
        atoms = Atoms(numbers=atom_numbers, scaled_positions=atom_xred, cell=atom_cell)
        m = MyTB(nbasis, nspin=nspin, ndim=ndim, positions=positions)
        m.atoms = copy.deepcopy(atoms)
        root.close()
        for iR, R in enumerate(Rlist):
            m.data[tuple(R)] = mdata_real[iR] + mdata_imag[iR] * 1j
        return m

    @staticmethod
    def from_tbmodel(model):
        """
        translate from a tbmodel type tight binding model
        """
        ret = MyTB(nbasis=model.size)
        for R, v in model.hop.items():
            ret.data[R] = v
        ret._positions = np.reshape(model.pos, (model.size, model.dim))
        return ret

    @staticmethod
    def from_tbmodel_hdf5(fname):
        """
        load model from a hdf5 file. It uses the tbmodel parser.
        """

        from tbmodels import Model

        m = Model.from_hdf5_file(fname)
        ret = MyTB(nbasis=m.size)
        for R, v in m.hop.items():
            ret.data[R] = v
        ret.positions = np.reshape(m.pos, (m.size, m.ndim))
        return ret

    def to_spin_polarized(self, order=1):
        """
        repeat to get spin polarized.
        order =1 : orb1_up, orb1_dn, orb2_up, orb2_dn...
        order =2 : orb1_up, orb2_up, ... orb1_dn, orb2_dn...
        """
        ret = MyTB(self.nbasis * 2)
        if self.positions is None:
            ret.positions = None
        else:
            ret.positions = np.repeat(self.positions, 2, axis=0)
        for R, mat in self.data.items():
            if order == 1:
                ret.data[R][::2, ::2] = mat
                ret.data[R][1::2, 1::2] = mat
            elif order == 2:
                ret.data[R][: self.norb, : self.norb] = mat
                ret.data[R][self.norb :, self.norb :] = mat
        return ret

    def validate(self):
        # make sure all R are 3d.
        for R in self.data.keys():
            if len(R) != self.ndim:
                raise ValueError("Dimension of R should be ndim %s" % (self.ndim))


def merge_tbmodels_spin(tbmodel_up, tbmodel_dn):
    """
    Merge a spin up and spin down model to one spinor model.
    """
    tbmodel = MyTB(
        nbasis=tbmodel_up.nbasis * 2,
        data=None,
        positions=np.vstack([tbmodel_up.positions, tbmodel_dn.positions]),
        sparse=False,
        ndim=tbmodel_up.ndim,
        nspin=2,
        double_site_energy=2.0,
    )
    norb = tbmodel.norb
    for R in tbmodel_up.data:
        tbmodel.data[R][:norb, :norb] = tbmodel_up.data[R][:, :]
        tbmodel.data[R][norb:, norb:] = tbmodel_dn.data[R][:, :]
    return tbmodel



#!/usr/bin/env python
"""
Parse the json file generated by the magnon band calculation. 
And plot the magnon band structure from the json file. 
You can modify this file to plot the magnon band structure, all plot 
multiple magnon band structures in one figure.
"""
import numpy as np
import matplotlib.pyplot as plt
import json


def load_magnon_json(filename):
    """
    Load the magnon band structure from the json file.
    args:
        filename: the json file generated by the magnon band calculation.
    return:
        kptlist: all the kpoints
        xlist: the x coordinates of the kpoints in the plot
        knames: names of the high symmetry points
        Xs: The x coordinates of the high symmetry points in the plot
        allevals: eigenvalues, in meV
            Note that the eigenvalues are splitted into groups of kpoints.
            the first index is the index of the group of kpoints
            the second index is the index of the kpoint in the group
            the third index is the index of the band
    """
    with open(filename) as f:
        data = json.load(f)
    kptlist = data["kptlist"]
    xlist = data["xlist"]
    knames = data["knames"]
    Xs = data["X_for_highsym_kpoints"]
    allevals = np.array(data["evals"])
    return kptlist, xlist, knames, Xs, allevals


def plot_magnon_json(filename, ax, color="k", show=True):
    """
    Plot the magnon band structure from the json file.
    args:
        filename: the json file generated by the magnon band calculation.
        ax: the matplotlib ax to plot the band structure.
            If None, a new figure will be created.
        color: the color of the band structure.
        show: whether to show the plot.
    """
    if ax is None:
        fig, ax = plt.subplots()

    kptlist, xlist, knames, Xs, allevals = load_magnon_json(filename)
    for i, (kpts, xs) in enumerate(zip(kptlist, xlist)):
        evals = allevals[i]
        # Plot band structure
        nbands = evals.shape[1]
        emin = np.min(evals[:, 0])
        for i in range(nbands):
            ax.plot(xs, (evals[:, i]) / 1.6e-22, color=color)

    ax.set_ylabel("Energy (meV)")
    ax.set_xlim(xlist[0][0], xlist[-1][-1])
    ax.set_xticks(Xs)
    knames = [x if x != "G" else "$\Gamma$" for x in knames]
    ax.set_xticklabels(knames)
    for x in Xs:
        ax.axvline(x, linewidth=0.6, color="gray")
    if show:
        plt.show()


if __name__ == "__main__":
    plot_magnon_json("magnon_band.json", ax=None, color="k", show=True)



from TB2J.exchange import ExchangeNCL
from TB2J.exchange_qspace import ExchangeCLQspace
from TB2J.exchangeCL2 import ExchangeCL2


class Manager:
    def __init__(self, atoms, models, basis, colinear, **kwargs):
        # computing exchange
        print("Starting to calculate exchange.")
        ExchangeClass = self.select_exchange(colinear)
        exchange = ExchangeClass(tbmodels=models, atoms=atoms, basis=basis, **kwargs)
        output_path = kwargs.get("output_path", "TB2J_results")
        exchange.run(path=output_path)
        print(f"All calculation finished. The results are in {output_path} directory.")

    def select_exchange(self, colinear, qspace=False):
        if colinear:
            if qspace:
                return ExchangeCLQspace
            else:
                return ExchangeCL2
        else:
            return ExchangeNCL
